from astromesh.observability.telemetry import TelemetryManager, TelemetryConfig, _NoOpSpan
from astromesh.observability.metrics import MetricsCollector, MetricsConfig
from astromesh.observability.cost_tracker import CostTracker, UsageRecord


def test_telemetry_disabled():
    config = TelemetryConfig(enabled=False)
    mgr = TelemetryManager(config)
    mgr.setup()
    assert mgr.get_tracer() is None


def test_telemetry_noop_span():
    span = _NoOpSpan()
    with span:
        span.set_attribute("key", "value")
        span.add_event("test")


def test_trace_agent_run_no_tracer():
    mgr = TelemetryManager(TelemetryConfig(enabled=False))
    mgr.setup()
    span = mgr.trace_agent_run("test-agent", "session1")
    assert isinstance(span, _NoOpSpan)


def test_metrics_disabled():
    config = MetricsConfig(enabled=False)
    collector = MetricsCollector(config)
    # Should not raise
    collector.record_agent_run("test", "react", "success", 0.5)
    collector.record_provider_call("ollama", "llama3", "success", 0.1)
    collector.record_tool_execution("search", "success")
    collector.record_tokens("test", 100, 50)


def test_cost_tracker_record():
    tracker = CostTracker()
    record = UsageRecord(
        agent_name="test", session_id="s1", model="llama3",
        provider="ollama", input_tokens=100, output_tokens=50,
        cost_usd=0.01, latency_ms=100, pattern="react",
    )
    tracker.record(record)
    assert tracker.get_total_cost() == 0.01


def test_cost_tracker_budget():
    tracker = CostTracker()
    tracker.set_budget("test", 1.0)
    tracker.record(UsageRecord(
        agent_name="test", session_id="s1", model="m", provider="p",
        input_tokens=100, output_tokens=50, cost_usd=0.5,
        latency_ms=100, pattern="react",
    ))
    budget = tracker.check_budget("test")
    assert budget["spent"] == 0.5
    assert budget["remaining"] == 0.5
    assert not budget["exceeded"]


def test_cost_tracker_exceeded():
    tracker = CostTracker()
    tracker.set_budget("test", 0.01)
    tracker.record(UsageRecord(
        agent_name="test", session_id="s1", model="m", provider="p",
        input_tokens=100, output_tokens=50, cost_usd=0.02,
        latency_ms=100, pattern="react",
    ))
    assert tracker.check_budget("test")["exceeded"]


def test_cost_tracker_summary():
    tracker = CostTracker()
    tracker.record(UsageRecord(agent_name="a", session_id="s1", model="m1",
        provider="p1", input_tokens=100, output_tokens=50, cost_usd=0.01,
        latency_ms=100, pattern="react"))
    tracker.record(UsageRecord(agent_name="a", session_id="s2", model="m2",
        provider="p2", input_tokens=200, output_tokens=100, cost_usd=0.02,
        latency_ms=200, pattern="react"))
    summary = tracker.get_usage_summary("a")
    assert summary["total_cost"] == 0.03
    assert summary["num_calls"] == 2
    assert "p1" in summary["by_provider"]
    assert "m2" in summary["by_model"]


def test_cost_tracker_filter_by_session():
    tracker = CostTracker()
    tracker.record(UsageRecord(agent_name="a", session_id="s1", model="m",
        provider="p", input_tokens=100, output_tokens=50, cost_usd=0.01,
        latency_ms=100, pattern="react"))
    tracker.record(UsageRecord(agent_name="a", session_id="s2", model="m",
        provider="p", input_tokens=200, output_tokens=100, cost_usd=0.02,
        latency_ms=200, pattern="react"))
    assert tracker.get_total_cost(session_id="s1") == 0.01



"""WindowsServiceManager — Windows Service adapter."""

from __future__ import annotations

import asyncio
import logging
from typing import Any, Callable

logger = logging.getLogger("astromesh_node.platform.windows")

SERVICE_NAME = "astromeshd"
SERVICE_DISPLAY = "Astromesh Agent Runtime Daemon"


class WindowsServiceManager:
    """ServiceManager implementation for Windows Services."""

    def __init__(self) -> None:
        self._reload_handler: Callable[[], Any] | None = None
        self._service_handle = None

    async def notify_ready(self) -> None:
        pass

    async def notify_reload(self) -> None:
        pass

    async def notify_stopping(self) -> None:
        pass

    async def install_service(self, profile: str) -> None:
        try:
            import win32serviceutil
            import win32service

            win32serviceutil.InstallService(
                None,
                SERVICE_NAME,
                SERVICE_DISPLAY,
                startType=win32service.SERVICE_AUTO_START,
            )
            logger.info("Installed Windows Service: %s", SERVICE_NAME)
        except ImportError:
            proc = await asyncio.create_subprocess_exec(
                "sc",
                "create",
                SERVICE_NAME,
                f"binPath={_get_service_binpath()}",
                "start=auto",
                f"DisplayName={SERVICE_DISPLAY}",
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            stdout, stderr = await proc.communicate()
            if proc.returncode == 0:
                logger.info("Installed Windows Service via sc.exe: %s", SERVICE_NAME)
            else:
                raise RuntimeError(f"Failed to install service: {stderr.decode()}")

    async def uninstall_service(self) -> None:
        try:
            import win32serviceutil

            win32serviceutil.RemoveService(SERVICE_NAME)
            logger.info("Removed Windows Service: %s", SERVICE_NAME)
        except ImportError:
            proc = await asyncio.create_subprocess_exec(
                "sc",
                "delete",
                SERVICE_NAME,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            await proc.communicate()

    async def service_status(self) -> dict[str, Any]:
        proc = await asyncio.create_subprocess_exec(
            "sc",
            "query",
            SERVICE_NAME,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        stdout_bytes, _ = await proc.communicate()
        stdout = stdout_bytes.decode()
        running = "RUNNING" in stdout if proc.returncode == 0 else False
        pid = None
        for line in stdout.splitlines():
            if "PID" in line:
                pid = int("".join(c for c in line if c.isdigit()) or "0") or None
                break

        return {
            "running": running,
            "enabled": proc.returncode == 0,
            "pid": pid,
            "mode": "windows_service",
        }

    def register_reload_handler(self, callback: Callable[[], Any]) -> None:
        self._reload_handler = callback


def _get_service_binpath() -> str:
    """Get the binpath for sc.exe create command."""
    import sys
    from pathlib import Path

    venv = Path(sys.executable).parent
    service_py = venv.parent / "packaging" / "windows" / "astromeshd-service.py"
    return f'"{sys.executable}" "{service_py}"'



import os

from astromesh.rag.chunking.base import ChunkingStrategy

try:
    from astromesh._native import rust_fixed_chunk as _native_chunk
except ImportError:
    _native_chunk = None


class FixedChunker(ChunkingStrategy):
    def __init__(self, chunk_size: int = 500, overlap: int = 50):
        self.chunk_size = chunk_size
        self.overlap = overlap

    def chunk(self, document: str, metadata: dict) -> list[dict]:
        if not document:
            return []

        if _native_chunk is not None and not os.environ.get("ASTROMESH_FORCE_PYTHON"):
            return _native_chunk(document, metadata, self.chunk_size, self.overlap)

        chunks: list[dict] = []
        start = 0
        i = 0

        while start < len(document):
            end = start + self.chunk_size
            content = document[start:end]
            chunks.append(
                {
                    "content": content,
                    "metadata": {**metadata, "chunk_index": i, "strategy": "fixed"},
                }
            )
            i += 1
            start += self.chunk_size - self.overlap
            if start >= len(document):
                break

        return chunks



"""Integration tests for Astromesh Node Phase 2 — services and peers."""

import pytest
from asgi_lifespan import LifespanManager
from httpx import ASGITransport, AsyncClient

from astromesh.api.main import app
from astromesh.api.routes import system
from astromesh.runtime.engine import AgentRuntime
from astromesh.runtime.peers import PeerClient
from astromesh.runtime.services import ServiceManager


@pytest.fixture(autouse=True)
def _skip_runtime_lifespan(monkeypatch):
    monkeypatch.setenv("ASTROMESH_SKIP_RUNTIME", "1")

AGENT_YAML = """
apiVersion: astromesh/v1
kind: Agent
metadata:
  name: test-agent
  version: "1.0.0"
  namespace: testing
spec:
  identity:
    display_name: Test Agent
    description: A test agent
  model:
    primary:
      provider: ollama
      model: test-model
      endpoint: http://localhost:11434
  prompts:
    system: "You are a test agent."
  orchestration:
    pattern: react
    max_iterations: 3
"""


@pytest.fixture
async def worker_node(tmp_path):
    agents_dir = tmp_path / "agents"
    agents_dir.mkdir()
    (agents_dir / "test.agent.yaml").write_text(AGENT_YAML)

    sm = ServiceManager({"agents": True, "inference": False, "channels": False})
    pc = PeerClient(
        [
            {"name": "inference-1", "url": "http://inference:8000", "services": ["inference"]},
        ]
    )

    runtime = AgentRuntime(config_dir=str(tmp_path), service_manager=sm, peer_client=pc)
    await runtime.bootstrap()
    system.set_runtime(runtime)

    async with LifespanManager(app):
        async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as client:
            yield client, runtime

    system.set_runtime(None)
    await pc.close()


async def test_worker_node_status(worker_node):
    client, runtime = worker_node
    resp = await client.get("/v1/system/status")
    assert resp.status_code == 200
    data = resp.json()

    assert data["agents_loaded"] == 1
    assert data["services"]["agents"] is True
    assert data["services"]["inference"] is False
    assert data["services"]["channels"] is False
    assert len(data["peers"]) == 1
    assert data["peers"][0]["name"] == "inference-1"


async def test_worker_node_doctor(worker_node):
    client, runtime = worker_node
    resp = await client.get("/v1/system/doctor")
    assert resp.status_code == 200
    data = resp.json()

    assert data["checks"]["runtime"]["status"] == "ok"
    assert "peer:inference-1" in data["checks"]
    assert data["checks"]["peer:inference-1"]["status"] == "unreachable"


@pytest.fixture
async def gateway_node(tmp_path):
    sm = ServiceManager(
        {"agents": False, "inference": False, "tools": False, "memory": False, "rag": False}
    )
    pc = PeerClient(
        [
            {
                "name": "worker-1",
                "url": "http://worker:8000",
                "services": ["agents", "tools", "memory"],
            },
        ]
    )

    runtime = AgentRuntime(config_dir=str(tmp_path), service_manager=sm, peer_client=pc)
    await runtime.bootstrap()
    system.set_runtime(runtime)

    async with LifespanManager(app):
        async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as client:
            yield client, runtime

    system.set_runtime(None)
    await pc.close()


async def test_gateway_node_no_agents(gateway_node):
    client, runtime = gateway_node
    resp = await client.get("/v1/system/status")
    assert resp.status_code == 200
    data = resp.json()
    assert data["agents_loaded"] == 0
    assert data["services"]["agents"] is False



import pytest
from astromesh.core.guardrails import GuardrailsEngine


async def test_pii_redaction_email():
    engine = GuardrailsEngine(config={
        "input": [{"type": "pii_detection", "action": "redact"}]
    })
    result = await engine.apply_input("My email is john@example.com")
    assert "john@example.com" not in result
    assert "[REDACTED_EMAIL]" in result


async def test_pii_redaction_phone():
    engine = GuardrailsEngine(config={
        "input": [{"type": "pii_detection", "action": "redact"}]
    })
    result = await engine.apply_input("Call me at 555-123-4567")
    assert "555-123-4567" not in result
    assert "[REDACTED_PHONE]" in result


async def test_topic_filter_warn():
    engine = GuardrailsEngine(config={
        "input": [{"type": "topic_filter", "blocked_topics": ["competitor"]}]
    })
    result = await engine.apply_input("Tell me about competitor products")
    assert result is not None


async def test_topic_filter_block():
    engine = GuardrailsEngine(config={
        "input": [{"type": "topic_filter", "blocked_topics": ["secret"], "action": "block"}]
    })
    with pytest.raises(ValueError, match="Blocked topic"):
        await engine.apply_input("Tell me the secret")


async def test_output_cost_limit():
    engine = GuardrailsEngine(config={
        "output": [{"type": "cost_limit", "max_tokens_per_turn": 10}]
    })
    long_text = " ".join(["word"] * 100)
    result = await engine.apply_output(long_text)
    assert len(result) < len(long_text)
    assert "[truncated]" in result


async def test_no_guardrails():
    engine = GuardrailsEngine()
    result = await engine.apply_input("hello")
    assert result == "hello"
    result = await engine.apply_output("world")
    assert result == "world"



"""Model router with strategy-based provider selection and circuit breaker."""

from __future__ import annotations

import os
import time

from astromesh.errors import (
    ModelProviderError,
    explain_model_provider_failure,
    explain_no_eligible_providers,
)
from astromesh.providers.base import (
    CompletionResponse,
    ProviderHealth,
    ProviderProtocol,
    RoutingStrategy,
)

try:
    from astromesh._native import rust_detect_vision, rust_ema_update

    _HAS_NATIVE_ROUTING = True
except ImportError:
    _HAS_NATIVE_ROUTING = False


class ModelRouter:
    """Routes completion requests across registered providers.

    Supports multiple routing strategies (cost, latency, round-robin,
    capability match) and implements a circuit breaker that opens after
    three consecutive failures with a 60-second cooldown.
    """

    CIRCUIT_BREAKER_THRESHOLD = 3
    CIRCUIT_BREAKER_COOLDOWN_S = 60.0
    EMA_ALPHA = 0.8  # weight for existing average
    EMA_BETA = 0.2  # weight for new observation

    def __init__(self, config: dict):
        self._providers: dict[str, ProviderProtocol] = {}
        self._health: dict[str, ProviderHealth] = {}
        self._strategy = RoutingStrategy(config.get("strategy", "cost_optimized"))
        self._config = config
        self._request_count = 0

    # ------------------------------------------------------------------
    # Public API
    # ------------------------------------------------------------------

    def register_provider(self, name: str, provider: ProviderProtocol) -> None:
        """Register a provider and initialise its health record."""
        self._providers[name] = provider
        self._health[name] = ProviderHealth()

    async def route(
        self,
        messages: list[dict],
        requirements: dict | None = None,
        **kwargs,
    ) -> CompletionResponse:
        """Route a completion request to the best available provider.

        Tries candidates in ranked order.  On failure the next candidate
        is attempted.  After *CIRCUIT_BREAKER_THRESHOLD* consecutive
        failures a provider's circuit is opened for *CIRCUIT_BREAKER_COOLDOWN_S*
        seconds.

        Raises ``ModelProviderError`` when every candidate has been exhausted
        or no providers are configured.
        """
        # Check for request-scoped provider override (BYOK)
        provider_override = kwargs.pop("provider_override", None)
        if provider_override:
            override_name, override_provider = provider_override
            try:
                response = await override_provider.complete(messages, **kwargs)
                response.latency_ms = 0.0
                return response
            except Exception as e:
                raise ModelProviderError(
                    f"Request-scoped provider '{override_name}' could not complete the request.",
                    hint=(
                        "Check X-Astromesh-Provider-Name / X-Astromesh-Provider-Key headers, "
                        "network access, and that the provider accepts the chosen model."
                    ),
                    code="model_provider_override_failed",
                    cause=e,
                ) from e

        if requirements is None:
            requirements = {}

        if not requirements.get("vision") and self._detect_vision_requirement(messages):
            requirements["vision"] = True

        registered = list(self._providers.keys())
        if not registered:
            raise explain_model_provider_failure(
                None,
                candidate_names=[],
                registered_provider_names=[],
            )

        candidates = self._rank_candidates(requirements)
        if not candidates:
            raise explain_no_eligible_providers(registered)

        last_error: Exception | None = None

        for name in candidates:
            provider = self._providers[name]
            health = self._health[name]

            start = time.perf_counter()
            try:
                response = await provider.complete(messages, **kwargs)
                elapsed_ms = (time.perf_counter() - start) * 1000.0

                # Update health metrics with exponential moving average.
                if health.avg_latency_ms == 0.0:
                    health.avg_latency_ms = elapsed_ms
                else:
                    if _HAS_NATIVE_ROUTING and not os.environ.get("ASTROMESH_FORCE_PYTHON"):
                        health.avg_latency_ms = rust_ema_update(
                            health.avg_latency_ms, elapsed_ms, self.EMA_ALPHA, self.EMA_BETA
                        )
                    else:
                        health.avg_latency_ms = (
                            self.EMA_ALPHA * health.avg_latency_ms + self.EMA_BETA * elapsed_ms
                        )
                health.consecutive_failures = 0
                health.is_healthy = True
                health.last_check = time.time()

                # Ensure latency_ms is set on the response.
                response.latency_ms = elapsed_ms

                self._request_count += 1
                return response

            except Exception as exc:  # noqa: BLE001
                last_error = exc
                health.consecutive_failures += 1
                health.last_check = time.time()

                if health.consecutive_failures >= self.CIRCUIT_BREAKER_THRESHOLD:
                    health.circuit_open = True
                    health.circuit_open_until = time.time() + self.CIRCUIT_BREAKER_COOLDOWN_S
                    health.is_healthy = False

        raise explain_model_provider_failure(
            last_error,
            candidate_names=candidates,
            registered_provider_names=registered,
        )

    @staticmethod
    def _detect_vision_requirement(messages: list[dict]) -> bool:
        """Return True if any message contains image_url content."""
        if _HAS_NATIVE_ROUTING and not os.environ.get("ASTROMESH_FORCE_PYTHON"):
            return rust_detect_vision(messages)
        for msg in messages:
            content = msg.get("content")
            if isinstance(content, list):
                for part in content:
                    if part.get("type") == "image_url":
                        return True
        return False

    # ------------------------------------------------------------------
    # Internal helpers
    # ------------------------------------------------------------------

    def _rank_candidates(self, requirements: dict) -> list[str]:
        """Return provider names ordered by the active routing strategy.

        Circuit-broken providers whose cooldown has not yet elapsed are
        excluded from the candidate list.
        """
        now = time.time()
        available: list[str] = []

        for name, health in self._health.items():
            if health.circuit_open:
                if now >= health.circuit_open_until:
                    # Cooldown elapsed – allow a retry (half-open).
                    health.circuit_open = False
                    health.consecutive_failures = 0
                    available.append(name)
                # else: still within cooldown, skip
            else:
                available.append(name)

        strategy = self._strategy

        if strategy == RoutingStrategy.COST_OPTIMIZED:
            available.sort(
                key=lambda n: self._providers[n].estimated_cost("default", 1000, 1000),
            )

        elif strategy == RoutingStrategy.LATENCY_OPTIMIZED:
            available.sort(key=lambda n: self._health[n].avg_latency_ms)

        elif strategy == RoutingStrategy.ROUND_ROBIN:
            if available:
                offset = self._request_count % len(available)
                available = available[offset:] + available[:offset]

        elif strategy == RoutingStrategy.CAPABILITY_MATCH:
            required_tools = requirements.get("tools", False)
            required_vision = requirements.get("vision", False)
            available = [
                n
                for n in available
                if (not required_tools or self._providers[n].supports_tools())
                and (not required_vision or self._providers[n].supports_vision())
            ]

        return available



from astromesh.runtime.engine import _parse_args, _truncate, _normalize_tool_calls


def test_truncate_none():
    assert _truncate(None, 100) == ""


def test_truncate_empty():
    assert _truncate("", 100) == ""


def test_truncate_within_limit():
    assert _truncate("hello", 100) == "hello"


def test_truncate_at_limit():
    text = "a" * 100
    assert _truncate(text, 100) == text


def test_truncate_over_limit():
    text = "a" * 200
    result = _truncate(text, 100)
    assert len(result) > 100
    assert result.startswith("a" * 100)
    assert "[truncated at 200 chars]" in result


def test_parse_args_dict_passthrough():
    assert _parse_args({"key": "val"}) == {"key": "val"}


def test_parse_args_json_string():
    assert _parse_args('{"key": "val"}') == {"key": "val"}


def test_parse_args_invalid_json():
    assert _parse_args("not json") == {"_raw": "not json"}


def test_normalize_flat_dict():
    raw = [{"id": "1", "name": "search", "arguments": {"q": "test"}}]
    assert _normalize_tool_calls(raw) == raw


def test_normalize_openai_nested():
    raw = [{"id": "call_1", "function": {"name": "search", "arguments": {"q": "test"}}}]
    result = _normalize_tool_calls(raw)
    assert result == [{"id": "call_1", "name": "search", "arguments": {"q": "test"}}]


def test_normalize_empty():
    assert _normalize_tool_calls([]) == []


def test_normalize_non_dict():
    result = _normalize_tool_calls(["not_a_dict"])
    assert result == [{"raw": "not_a_dict"}]


def test_normalize_json_string_args():
    raw = [{"id": "call_1", "function": {"name": "search", "arguments": '{"q": "test"}'}}]
    result = _normalize_tool_calls(raw)
    assert result == [{"id": "call_1", "name": "search", "arguments": {"q": "test"}}]



from __future__ import annotations

import httpx

from astromesh.rag.embeddings.base import EmbeddingProvider

_DEFAULT_HF_API = "https://api-inference.huggingface.co/pipeline/feature-extraction"


class HFEmbeddingProvider(EmbeddingProvider):
    """Hugging Face Inference API / TEI embedding provider."""

    def __init__(
        self,
        model_name: str = "BAAI/bge-small-en-v1.5",
        endpoint: str | None = None,
        api_key: str | None = None,
    ):
        self.model_name = model_name
        self.endpoint = endpoint or f"{_DEFAULT_HF_API}/{model_name}"
        self.api_key = api_key

    def _headers(self) -> dict[str, str]:
        headers: dict[str, str] = {"Content-Type": "application/json"}
        if self.api_key:
            headers["Authorization"] = f"Bearer {self.api_key}"
        return headers

    async def embed(self, text: str) -> list[float]:
        async with httpx.AsyncClient() as client:
            response = await client.post(
                self.endpoint,
                json={"inputs": text},
                headers=self._headers(),
                timeout=30.0,
            )
            response.raise_for_status()
            data = response.json()
            # HF API may return nested list or flat list
            if isinstance(data, list) and data and isinstance(data[0], list):
                return data[0]
            return data

    async def embed_batch(self, texts: list[str]) -> list[list[float]]:
        async with httpx.AsyncClient() as client:
            response = await client.post(
                self.endpoint,
                json={"inputs": texts},
                headers=self._headers(),
                timeout=60.0,
            )
            response.raise_for_status()
            return response.json()



"""Orbit CLI — registers as astromeshctl plugin."""

from __future__ import annotations

import asyncio
from pathlib import Path
from typing import Optional

import typer
from rich.console import Console
from rich.table import Table

from astromesh_orbit.config import OrbitConfig
from astromesh_orbit.providers.gcp.provider import GCPProvider
from astromesh_orbit.terraform.runner import TerraformNotFoundError
from astromesh_orbit.wizard.interactive import run_wizard

console = Console()
orbit_app = typer.Typer(
    name="orbit", help="Cloud-native deployment for Astromesh.", no_args_is_help=True
)

PROVIDERS = {"gcp": GCPProvider}
ORBIT_DIR = Path(".orbit")
GENERATED_DIR = ORBIT_DIR / "generated"


def _load_config(config_path: str) -> OrbitConfig:
    path = Path(config_path)
    if not path.exists():
        console.print(f"[red]Error:[/] {path} not found. Run 'astromeshctl orbit init' first.")
        raise typer.Exit(1)
    return OrbitConfig.from_yaml(path)


def _get_provider(config: OrbitConfig):
    name = config.spec.provider.name
    if name not in PROVIDERS:
        console.print(
            f"[red]Error:[/] Provider '{name}' not supported. Available: {list(PROVIDERS.keys())}"
        )
        raise typer.Exit(1)
    return PROVIDERS[name]()


@orbit_app.command()
def init(
    provider: str = typer.Option("gcp", help="Cloud provider"),
    preset: Optional[str] = typer.Option(None, help="Preset: starter or pro"),
):
    """Interactive setup — generates orbit.yaml."""
    run_wizard()


@orbit_app.command()
def plan(config: str = typer.Option("orbit.yaml", help="Path to orbit.yaml")):
    """Preview infrastructure changes."""

    async def _plan():
        cfg = _load_config(config)
        prov = _get_provider(cfg)

        console.print("\n  [cyan bold]Orbit Deployment Plan[/]\n")

        # Validate
        console.print("  Validating...", end="")
        validation = await prov.validate(cfg)
        if not validation.ok:
            console.print(" [red]FAILED[/]\n")
            for c in validation.checks:
                icon = "[green]OK[/]" if c.passed else "[red]FAIL[/]"
                console.print(f"    {icon} {c.message}")
                if c.remediation:
                    console.print(f"      -> {c.remediation}")
            raise typer.Exit(1)
        console.print(" [green]OK[/]")

        # Generate and plan
        GENERATED_DIR.mkdir(parents=True, exist_ok=True)
        await prov.generate(cfg, GENERATED_DIR)
        try:
            from astromesh_orbit.terraform.runner import TerraformRunner

            runner = TerraformRunner()
            await runner.check_installed()
            from astromesh_orbit.terraform.backend import ensure_gcs_state_bucket

            await ensure_gcs_state_bucket(
                cfg.spec.provider.project, cfg.spec.provider.region, cfg.metadata.name
            )
            await runner.init(GENERATED_DIR)
            result = await runner.plan(GENERATED_DIR)
            console.print(f"\n  Resources to create: {len(result.resources_to_create)}")
            console.print(f"  Resources to update: {len(result.resources_to_update)}")
            console.print(f"  Resources to destroy: {len(result.resources_to_destroy)}")
            if result.estimated_monthly_cost:
                console.print(f"  Estimated cost: ~${result.estimated_monthly_cost}/month")
        except TerraformNotFoundError as e:
            console.print(f"\n  [red]{e}[/]")
            raise typer.Exit(1)

    asyncio.run(_plan())


@orbit_app.command()
def apply(
    config: str = typer.Option("orbit.yaml", help="Path to orbit.yaml"),
    auto_approve: bool = typer.Option(False, "--auto-approve", help="Skip confirmation"),
):
    """Deploy infrastructure to the cloud."""

    async def _apply():
        cfg = _load_config(config)
        prov = _get_provider(cfg)
        GENERATED_DIR.mkdir(parents=True, exist_ok=True)

        console.print("\n  [cyan bold]Astromesh Orbit -- Deploying[/]\n")

        try:
            result = await prov.provision(cfg, GENERATED_DIR)
        except RuntimeError as e:
            console.print(f"  [red]Error:[/] {e}")
            raise typer.Exit(1)

        console.print("\n  [green bold]OK Deployment complete![/]\n")

        table = Table(title="Endpoints")
        table.add_column("Service", style="cyan")
        table.add_column("URL", style="green")
        for svc, url in result.endpoints.items():
            if url:
                table.add_row(svc, url)
        console.print(table)
        console.print(f"\n  Environment file: {result.env_file}\n")

    asyncio.run(_apply())


@orbit_app.command()
def status(config: str = typer.Option("orbit.yaml", help="Path to orbit.yaml")):
    """Show deployment status."""

    async def _status():
        cfg = _load_config(config)
        prov = _get_provider(cfg)
        ds = await prov.status(cfg)

        table = Table(title="Deployment Status")
        table.add_column("Resource", style="cyan")
        table.add_column("Type")
        table.add_column("Status")
        table.add_column("URL", style="dim")
        for r in ds.resources:
            color = "green" if r.status == "running" else "red"
            table.add_row(r.name, r.resource_type, f"[{color}]{r.status}[/]", r.url or "--")
        console.print(table)
        console.print(f"\n  State bucket: {ds.state_bucket}")

    asyncio.run(_status())


@orbit_app.command()
def destroy(
    config: str = typer.Option("orbit.yaml", help="Path to orbit.yaml"),
    auto_approve: bool = typer.Option(False, "--auto-approve", help="Skip confirmation"),
):
    """Destroy all provisioned infrastructure."""

    async def _destroy():
        cfg = _load_config(config)
        prov = _get_provider(cfg)

        if not auto_approve:
            typer.confirm("This will destroy ALL infrastructure. Continue?", abort=True)

        console.print("\n  [yellow]Destroying infrastructure...[/]\n")
        await prov.destroy(cfg, GENERATED_DIR)
        console.print("  [green]OK All resources destroyed.[/]\n")

    asyncio.run(_destroy())


@orbit_app.command()
def eject(output_dir: str = typer.Option("./orbit-terraform", help="Output directory")):
    """Export standalone Terraform files."""

    async def _eject():
        cfg = _load_config("orbit.yaml")
        prov = _get_provider(cfg)
        result = await prov.eject(cfg, Path(output_dir))
        console.print(f"\n  [green]OK[/] Terraform files exported to {result}/")
        console.print("  These are standalone -- no Orbit dependency.\n")
        console.print("  Next steps:")
        console.print(f"    cd {output_dir}")
        console.print("    terraform plan")
        console.print("    terraform apply\n")

    asyncio.run(_eject())


def register(app: typer.Typer) -> None:
    """Plugin entry point — called by astromeshctl plugin discovery."""
    app.add_typer(orbit_app, name="orbit")



"""Tests for mesh API endpoints."""

import time

import pytest
from fastapi.testclient import TestClient

from astromesh.api.main import app
from astromesh.api.routes import mesh as mesh_routes
from astromesh.mesh.config import MeshConfig
from astromesh.mesh.leader import LeaderElector
from astromesh.mesh.manager import MeshManager
from astromesh.runtime.services import ServiceManager


def _make_node(node_id: str = "node-2", name: str = "peer") -> dict:
    return {
        "node_id": node_id,
        "name": name,
        "url": "http://127.0.0.1:9000",
        "services": ["api"],
        "agents": [],
        "load": {"cpu_percent": 10.0, "memory_percent": 20.0, "active_requests": 0},
        "leader": False,
        "joined_at": time.time(),
        "last_heartbeat": time.time(),
        "status": "alive",
    }


@pytest.fixture()
def mesh_manager():
    config = MeshConfig(enabled=True, node_name="test-node", bind="127.0.0.1:8000")
    svc = ServiceManager({"api": True})
    return MeshManager(config, svc)


@pytest.fixture()
def elector(mesh_manager):
    return LeaderElector(mesh_manager)


@pytest.fixture()
def client(mesh_manager, elector):
    mesh_routes.set_mesh(mesh_manager, elector)
    yield TestClient(app)
    mesh_routes.set_mesh(None, None)


def test_mesh_state(client, mesh_manager):
    resp = client.get("/v1/mesh/state")
    assert resp.status_code == 200
    data = resp.json()
    assert "nodes" in data
    assert "leader_id" in data
    assert len(data["nodes"]) == 1
    assert data["nodes"][0]["node_id"] == mesh_manager.node_id


def test_mesh_join(client, mesh_manager):
    node_data = _make_node()
    resp = client.post("/v1/mesh/join", json=node_data)
    assert resp.status_code == 200
    data = resp.json()
    assert len(data["nodes"]) == 2
    node_ids = [n["node_id"] for n in data["nodes"]]
    assert "node-2" in node_ids
    assert mesh_manager.node_id in node_ids


def test_mesh_leave(client, mesh_manager):
    # First join a node
    node_data = _make_node()
    client.post("/v1/mesh/join", json=node_data)
    assert len(mesh_manager.cluster_state().nodes) == 2

    # Then leave
    resp = client.post("/v1/mesh/leave", json={"node_id": "node-2"})
    assert resp.status_code == 200
    assert resp.json() == {"status": "ok"}
    assert len(mesh_manager.cluster_state().nodes) == 1


def test_mesh_heartbeat(client, mesh_manager):
    node_data = _make_node()
    resp = client.post("/v1/mesh/heartbeat", json=node_data)
    assert resp.status_code == 200
    assert resp.json() == {"status": "ok"}
    assert "node-2" in mesh_manager.cluster_state().nodes


def test_mesh_gossip(client, mesh_manager):
    nodes = [_make_node("node-3", "gossip-peer")]
    resp = client.post("/v1/mesh/gossip", json={"nodes": nodes})
    assert resp.status_code == 200
    data = resp.json()
    returned_ids = [n["node_id"] for n in data["nodes"]]
    assert "node-3" in returned_ids
    assert mesh_manager.node_id in returned_ids


def test_mesh_election(client, mesh_manager):
    resp = client.post(
        "/v1/mesh/election",
        json={"candidate_id": mesh_manager.node_id, "node_id": mesh_manager.node_id},
    )
    assert resp.status_code == 200
    data = resp.json()
    assert "leader_id" in data
    assert data["leader_id"] == mesh_manager.node_id


def test_mesh_state_not_enabled():
    mesh_routes.set_mesh(None, None)
    with TestClient(app) as c:
        resp = c.get("/v1/mesh/state")
        assert resp.status_code == 503
        assert "Mesh not enabled" in resp.json()["detail"]



"""Tests for ForegroundManager (no-op adapter)."""

import pytest

from astromesh_node.platform.base import ServiceManagerProtocol
from astromesh_node.platform.foreground import ForegroundManager


@pytest.fixture
def manager():
    return ForegroundManager()


def test_implements_protocol(manager):
    assert isinstance(manager, ServiceManagerProtocol)


async def test_notify_ready_is_noop(manager):
    await manager.notify_ready()  # Should not raise


async def test_notify_reload_is_noop(manager):
    await manager.notify_reload()


async def test_notify_stopping_is_noop(manager):
    await manager.notify_stopping()


async def test_service_status_returns_foreground(manager):
    status = await manager.service_status()
    assert status["running"] is True
    assert status["mode"] == "foreground"


def test_register_reload_handler_stores_callback(manager):
    called = []
    manager.register_reload_handler(lambda: called.append(True))


async def test_install_service_raises_in_foreground(manager):
    with pytest.raises(RuntimeError, match="foreground"):
        await manager.install_service("full")


async def test_uninstall_service_raises_in_foreground(manager):
    with pytest.raises(RuntimeError, match="foreground"):
        await manager.uninstall_service()



"""State bucket management for Terraform remote backends."""

from __future__ import annotations

import asyncio
import hashlib

from rich.console import Console

console = Console()


async def ensure_gcs_state_bucket(project: str, region: str, name: str) -> str:
    """Create GCS bucket for Terraform state if it doesn't exist. Returns bucket name."""
    bucket_name = f"{project}-astromesh-orbit-state"

    # Check if bucket exists
    proc = await asyncio.create_subprocess_exec(
        "gsutil",
        "ls",
        f"gs://{bucket_name}",
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    await proc.communicate()

    if proc.returncode == 0:
        console.print(f"  [green]OK[/] State bucket exists: gs://{bucket_name}")
        return bucket_name

    # Try to create
    proc = await asyncio.create_subprocess_exec(
        "gsutil",
        "mb",
        "-p",
        project,
        "-l",
        region,
        "-b",
        "on",
        f"gs://{bucket_name}",
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    _, stderr = await proc.communicate()

    if proc.returncode == 0:
        # Enable versioning
        await asyncio.create_subprocess_exec(
            "gsutil",
            "versioning",
            "set",
            "on",
            f"gs://{bucket_name}",
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        console.print(f"  [green]OK[/] State bucket created: gs://{bucket_name}")
        return bucket_name

    # Naming collision — append hash
    suffix = hashlib.sha256(f"{project}-{name}".encode()).hexdigest()[:6]
    bucket_name = f"{project}-astromesh-orbit-state-{suffix}"
    proc = await asyncio.create_subprocess_exec(
        "gsutil",
        "mb",
        "-p",
        project,
        "-l",
        region,
        "-b",
        "on",
        f"gs://{bucket_name}",
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    _, stderr = await proc.communicate()

    if proc.returncode != 0:
        raise RuntimeError(
            f"Failed to create state bucket gs://{bucket_name}.\n"
            f"Grant roles/storage.admin or create the bucket manually:\n"
            f"  gsutil mb -p {project} -l {region} gs://{bucket_name}"
        )

    console.print(f"  [green]OK[/] State bucket created: gs://{bucket_name}")
    return bucket_name



"""HTTP client for communicating with astromeshd."""

import os

import httpx

DEFAULT_URL = "http://localhost:8000"


def get_base_url() -> str:
    return os.environ.get("ASTROMESH_DAEMON_URL", DEFAULT_URL)


def api_get(path: str) -> dict:
    url = f"{get_base_url()}{path}"
    resp = httpx.get(url, timeout=5.0)
    resp.raise_for_status()
    return resp.json()


def api_post(path: str, json: dict | None = None) -> dict:
    url = f"{get_base_url()}{path}"
    resp = httpx.post(url, json=json, timeout=5.0)
    resp.raise_for_status()
    return resp.json()


def api_post_with_timeout(path: str, json: dict | None = None, timeout: float = 30.0) -> dict:
    """POST with configurable timeout for long-running operations."""
    url = f"{get_base_url()}{path}"
    resp = httpx.post(url, json=json, timeout=timeout)
    resp.raise_for_status()
    return resp.json()


def api_get_params(path: str, params: dict | None = None, timeout: float = 5.0) -> dict:
    """GET with query parameters."""
    url = f"{get_base_url()}{path}"
    resp = httpx.get(url, params=params, timeout=timeout)
    resp.raise_for_status()
    return resp.json()



from fastapi import APIRouter, WebSocket, WebSocketDisconnect
import json

router = APIRouter()


class ConnectionManager:
    def __init__(self):
        self.active_connections: dict[str, list[WebSocket]] = {}

    async def connect(self, websocket: WebSocket, session_id: str):
        await websocket.accept()
        if session_id not in self.active_connections:
            self.active_connections[session_id] = []
        self.active_connections[session_id].append(websocket)

    def disconnect(self, websocket: WebSocket, session_id: str):
        if session_id in self.active_connections:
            self.active_connections[session_id].remove(websocket)
            if not self.active_connections[session_id]:
                del self.active_connections[session_id]

    async def send_message(self, message: dict, websocket: WebSocket):
        await websocket.send_json(message)


manager = ConnectionManager()


@router.websocket("/ws/agent/{agent_name}")
async def agent_websocket(websocket: WebSocket, agent_name: str):
    session_id = websocket.query_params.get("session_id", "default")
    await manager.connect(websocket, session_id)
    try:
        while True:
            data = await websocket.receive_text()
            message = json.loads(data)
            query = message.get("query", "")

            await manager.send_message(
                {
                    "type": "status",
                    "status": "processing",
                    "agent": agent_name,
                },
                websocket,
            )

            # Placeholder response — will be wired to runtime
            await manager.send_message(
                {
                    "type": "response",
                    "agent": agent_name,
                    "answer": f"[WebSocket] Received: {query}",
                    "session_id": session_id,
                },
                websocket,
            )
    except WebSocketDisconnect:
        manager.disconnect(websocket, session_id)



from fastapi import APIRouter
from pydantic import BaseModel

router = APIRouter()


class RAGIngestRequest(BaseModel):
    document: str
    metadata: dict = {}
    doc_id_prefix: str = "doc"


class RAGQueryRequest(BaseModel):
    query: str
    top_k: int = 5


@router.post("/rag/ingest")
async def ingest_document(request: RAGIngestRequest):
    return {"status": "not_configured", "chunks": 0}


@router.post("/rag/query")
async def query_rag(request: RAGQueryRequest):
    return {"query": request.query, "results": [], "top_k": request.top_k}



from urllib.parse import urlparse

import httpx

from astromesh.tools.base import BuiltinTool, ToolContext, ToolResult

_BLOCKED_HOSTS = {"localhost", "127.0.0.1", "0.0.0.0", "::1", "[::1]"}


class HttpRequestTool(BuiltinTool):
    name = "http_request"
    description = "Make HTTP requests (GET, POST, PUT, DELETE) to external APIs"
    parameters = {
        "type": "object",
        "properties": {
            "method": {
                "type": "string",
                "enum": ["GET", "POST", "PUT", "DELETE", "PATCH"],
            },
            "url": {"type": "string", "description": "Request URL"},
            "headers": {"type": "object", "description": "Request headers"},
            "body": {"description": "Request body (for POST/PUT/PATCH)"},
        },
        "required": ["method", "url"],
    }

    async def execute(self, arguments: dict, context: ToolContext) -> ToolResult:
        url = arguments["url"]
        method = arguments["method"].upper()
        headers = arguments.get("headers", {})
        body = arguments.get("body")
        allow_localhost = self.config.get("allow_localhost", False)
        if not allow_localhost:
            parsed = urlparse(url)
            if parsed.hostname in _BLOCKED_HOSTS:
                return ToolResult(
                    success=False,
                    data=None,
                    error="Blocked: requests to localhost are not allowed",
                )
        timeout = self.config.get("timeout_seconds", 30)
        max_size = self.config.get("max_response_bytes", 5 * 1024 * 1024)
        try:
            async with httpx.AsyncClient(timeout=timeout) as client:
                kwargs: dict = {"headers": headers}
                if body is not None and method in ("POST", "PUT", "PATCH"):
                    kwargs["json"] = body
                resp = await client.request(method, url, **kwargs)
                try:
                    resp_body = resp.json()
                except Exception:
                    resp_body = resp.text[:max_size]
                return ToolResult(
                    success=True,
                    data={
                        "status_code": resp.status_code,
                        "headers": dict(resp.headers),
                        "body": resp_body,
                    },
                    metadata={"url": url, "method": method},
                )
        except Exception as e:
            return ToolResult(success=False, data=None, error=str(e))


class GraphQLQueryTool(BuiltinTool):
    name = "graphql_query"
    description = "Execute GraphQL queries against an endpoint"
    parameters = {
        "type": "object",
        "properties": {
            "endpoint": {"type": "string"},
            "query": {"type": "string"},
            "variables": {"type": "object"},
            "headers": {"type": "object"},
        },
        "required": ["endpoint", "query"],
    }

    async def execute(self, arguments: dict, context: ToolContext) -> ToolResult:
        endpoint = arguments["endpoint"]
        query = arguments["query"]
        variables = arguments.get("variables", {})
        headers = arguments.get("headers", {})
        timeout = self.config.get("timeout_seconds", 30)
        try:
            async with httpx.AsyncClient(timeout=timeout) as client:
                resp = await client.post(
                    endpoint,
                    json={"query": query, "variables": variables},
                    headers=headers,
                )
                return ToolResult(
                    success=True,
                    data=resp.json(),
                    metadata={"endpoint": endpoint, "status_code": resp.status_code},
                )
        except Exception as e:
            return ToolResult(success=False, data=None, error=str(e))



"""astromeshd — Astromesh Agent Runtime Daemon (platform-agnostic)."""

from __future__ import annotations

import argparse
import asyncio
import logging
import os
import signal
import sys
from pathlib import Path

from astromesh_node.daemon.config import DaemonConfig, detect_config_dir
from astromesh_node.installer.detect import get_installer
from astromesh_node.platform.detect import get_service_manager

logger = logging.getLogger("astromeshd")


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog="astromeshd",
        description="Astromesh Agent Runtime Daemon",
    )
    parser.add_argument("--config", type=str, default=None, help="Config directory")
    parser.add_argument("--host", type=str, default=None, help="Bind host")
    parser.add_argument("--port", type=int, default=None, help="Bind port")
    parser.add_argument(
        "--log-level",
        type=str,
        default="info",
        choices=["debug", "info", "warning", "error"],
    )
    parser.add_argument(
        "--foreground",
        action="store_true",
        help="Run in foreground mode (no init system integration)",
    )
    parser.add_argument("--pid-file", type=str, default=None, help="PID file path")
    return parser.parse_args(argv)


def setup_logging(level: str) -> None:
    logging.basicConfig(
        level=getattr(logging, level.upper()),
        format="%(asctime)s [%(levelname)s] %(name)s: %(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S",
    )


def _default_pid_file() -> str:
    """Return the default PID file path for the current platform."""
    try:
        installer = get_installer()
        return str(installer.data_dir() / "astromeshd.pid")
    except Exception:
        return "/var/lib/astromesh/data/astromeshd.pid"


def write_pid_file(pid_file: str) -> None:
    Path(pid_file).parent.mkdir(parents=True, exist_ok=True)
    Path(pid_file).write_text(str(os.getpid()))


def remove_pid_file(pid_file: str) -> None:
    path = Path(pid_file)
    if path.exists():
        path.unlink()


async def run_daemon(args: argparse.Namespace) -> None:
    import uvicorn

    from astromesh.api.main import app
    from astromesh.api.routes import agents, system
    from astromesh.runtime.engine import AgentRuntime
    from astromesh.runtime.peers import PeerClient
    from astromesh.runtime.services import ServiceManager

    config_dir = detect_config_dir(args.config)
    daemon_config = DaemonConfig.from_config_dir(config_dir)

    host = args.host or daemon_config.host
    port = args.port or daemon_config.port
    pid_file = args.pid_file or _default_pid_file()

    # Platform service manager
    service_mgr = get_service_manager(foreground=args.foreground)

    # Create service manager and peer client
    service_manager = ServiceManager(daemon_config.services)
    peer_client = PeerClient(daemon_config.peers)

    # Create mesh if enabled
    mesh_manager = None
    elector = None
    from astromesh.mesh.config import MeshConfig

    mesh_config = MeshConfig.from_dict(daemon_config.mesh)
    if mesh_config.enabled:
        from astromesh.mesh.leader import LeaderElector
        from astromesh.mesh.manager import MeshManager

        mesh_manager = MeshManager(mesh_config, service_manager)
        elector = LeaderElector(mesh_manager)
        peer_client = PeerClient.from_mesh(mesh_manager)
        logger.info("Mesh enabled, node: %s", mesh_config.node_name)
        if daemon_config.peers:
            logger.warning("spec.peers ignored when mesh is enabled")

    enabled = service_manager.enabled_services()
    logger.info("Enabled services: %s", ", ".join(enabled))
    for warning in service_manager.validate():
        logger.warning("Config warning: %s", warning)
    if daemon_config.peers:
        logger.info("Peers: %s", ", ".join(p["name"] for p in daemon_config.peers))

    installer = get_installer()
    system_dir = str(installer.config_dir())
    mode = "system" if config_dir == system_dir else "dev"
    logger.info("Starting astromeshd in %s mode", mode)
    logger.info("Config directory: %s", config_dir)

    write_pid_file(pid_file)

    runtime = AgentRuntime(
        config_dir=config_dir,
        service_manager=service_manager,
        peer_client=peer_client,
    )
    await runtime.bootstrap()

    agents.set_runtime(runtime)
    system.set_runtime(runtime)

    from astromesh.api.routes import memory as memory_routes
    from astromesh.api.routes import templates as templates_route

    memory_routes.set_runtime(runtime)

    templates_path = Path(config_dir) / "templates"
    if templates_path.is_dir() and any(templates_path.glob("*.template.yaml")):
        templates_route.set_templates_dir(str(templates_path))

    from astromesh.api.routes import mesh as mesh_routes

    mesh_routes.set_mesh(mesh_manager, elector)

    agent_count = len(runtime.list_agents())
    logger.info("Loaded %d agent(s)", agent_count)

    if mesh_manager:
        mesh_manager.update_agents([a["name"] for a in runtime.list_agents()])
        await mesh_manager.join()
        elector.elect()
        logger.info("Mesh joined, cluster size: %d", len(mesh_manager.cluster_state().nodes))

    runtime.mesh_manager = mesh_manager

    await service_mgr.notify_ready()

    config = uvicorn.Config(
        app=app,
        host=host,
        port=port,
        log_level=args.log_level,
        access_log=True,
    )
    server = uvicorn.Server(config)

    loop = asyncio.get_event_loop()

    def handle_shutdown(sig, frame):
        logger.info("Received %s, shutting down...", signal.Signals(sig).name)
        server.should_exit = True

    if sys.platform != "win32":
        loop.add_signal_handler(signal.SIGTERM, lambda: handle_shutdown(signal.SIGTERM, None))
        loop.add_signal_handler(signal.SIGINT, lambda: handle_shutdown(signal.SIGINT, None))
    else:
        signal.signal(signal.SIGTERM, handle_shutdown)
        signal.signal(signal.SIGINT, handle_shutdown)

    def _reload_config():
        logger.info("Config reload triggered")

    service_mgr.register_reload_handler(_reload_config)

    mesh_tasks = []
    if mesh_manager:

        async def _gossip_loop():
            while not server.should_exit:
                try:
                    await mesh_manager.gossip_once()
                except Exception as e:
                    logger.debug("Gossip error: %s", e)
                await asyncio.sleep(mesh_config.gossip_interval)

        async def _heartbeat_loop():
            while not server.should_exit:
                try:
                    await mesh_manager.heartbeat_once()
                except Exception as e:
                    logger.debug("Heartbeat error: %s", e)
                await asyncio.sleep(mesh_config.heartbeat_interval)

        mesh_tasks.append(asyncio.create_task(_gossip_loop()))
        mesh_tasks.append(asyncio.create_task(_heartbeat_loop()))

    try:
        await server.serve()
    finally:
        await service_mgr.notify_stopping()
        for task in mesh_tasks:
            task.cancel()
        if mesh_manager:
            await mesh_manager.leave()
            await mesh_manager.close()
        remove_pid_file(pid_file)
        logger.info("astromeshd stopped")


def main() -> None:
    args = parse_args()
    setup_logging(args.log_level)
    asyncio.run(run_daemon(args))


if __name__ == "__main__":
    main()



import logging
import os
from contextlib import asynccontextmanager
from pathlib import Path

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles

from astromesh import __version__
from astromesh.api import ws
from astromesh.api.routes import (
    agents,
    dashboard,
    memory,
    tools,
    rag,
    whatsapp,
    system,
    mesh,
    traces,
    metrics,
    workflows,
    templates,
)
from astromesh.logging_config import setup_logging

setup_logging()

logger = logging.getLogger(__name__)


@asynccontextmanager
async def lifespan(app: FastAPI):
    """Bootstrap AgentRuntime for `uvicorn astromesh.api.main:app` (astromeshd sets runtime before serve)."""
    skip = os.environ.get("ASTROMESH_SKIP_RUNTIME", "").lower() in ("1", "true", "yes")
    if skip:
        yield
        return

    from astromesh.api.routes import agents as agents_route
    from astromesh.api.routes import memory as memory_route
    from astromesh.api.routes import system as system_route
    from astromesh.api.routes import whatsapp as whatsapp_route
    from astromesh.runtime.engine import AgentRuntime

    if agents_route._runtime is not None:
        # Runtime was injected before serve (e.g. astromeshd). Keep other route modules
        # in sync — tests that temporarily replace agents._runtime can otherwise leave
        # memory/system/whatsapp unwired on the next lifespan if teardown order restores
        # agents._runtime after the prior finally cleared the other routes.
        r = agents_route._runtime
        system_route.set_runtime(r)
        memory_route.set_runtime(r)
        whatsapp_route.set_runtime(r)
        yield
        return

    config_dir = os.environ.get("ASTROMESH_CONFIG_DIR", "config")
    runtime = AgentRuntime(config_dir=config_dir)
    try:
        await runtime.bootstrap()
    except Exception:
        logger.exception("AgentRuntime bootstrap failed (config_dir=%s)", config_dir)
        raise

    n_agents = len(runtime.list_agents())
    logger.info(
        "AgentRuntime ready: config_dir=%s agents_loaded=%d",
        config_dir,
        n_agents,
    )

    agents_route.set_runtime(runtime)
    system_route.set_runtime(runtime)
    memory_route.set_runtime(runtime)
    whatsapp_route.set_runtime(runtime)

    try:
        yield
    finally:
        agents_route.set_runtime(None)
        system_route.set_runtime(None)
        memory_route.set_runtime(None)
        whatsapp_route.set_runtime(None)


app = FastAPI(
    title="Astromesh Agent Runtime API",
    version=__version__,
    lifespan=lifespan,
)

# CORS for standalone Forge
cors_origins = os.getenv("ASTROMESH_CORS_ORIGINS", "http://localhost:5173").split(",")
app.add_middleware(
    CORSMiddleware,
    allow_origins=cors_origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

app.include_router(agents.router, prefix="/v1")
app.include_router(memory.router, prefix="/v1")
app.include_router(tools.router, prefix="/v1")
app.include_router(rag.router, prefix="/v1")
app.include_router(ws.router, prefix="/v1")
app.include_router(whatsapp.router, prefix="/v1")
app.include_router(system.router, prefix="/v1")
app.include_router(mesh.router, prefix="/v1")
app.include_router(traces.router, prefix="/v1")
app.include_router(metrics.router, prefix="/v1")
app.include_router(workflows.router, prefix="/v1")
app.include_router(dashboard.router, prefix="/v1")
app.include_router(templates.router, prefix="/v1")

# Optional Forge static files (embedded mode)
forge_static = Path(__file__).parent.parent / "static" / "forge"
if forge_static.exists():
    app.mount("/forge", StaticFiles(directory=str(forge_static), html=True), name="forge")


@app.get("/v1/health")
async def health():
    return {"status": "ok", "version": __version__}



"""astromeshctl agents commands."""

import typer
from rich.table import Table

from astromesh_cli.client import api_get
from astromesh_cli.output import console, print_error, print_json

app = typer.Typer(help="Manage agents.")


@app.command("list")
def list_agents(json: bool = typer.Option(False, "--json", help="Output as JSON")):
    """List loaded agents."""
    try:
        data = api_get("/v1/agents")
        if json:
            print_json(data)
            return

        agents = data.get("agents", [])
        if not agents:
            console.print("[dim]No agents loaded.[/dim]")
            return

        table = Table(title="Loaded Agents")
        table.add_column("Name", style="cyan")
        table.add_column("Version", style="green")
        table.add_column("Namespace", style="dim")

        for agent in agents:
            table.add_row(
                agent.get("name", ""),
                agent.get("version", ""),
                agent.get("namespace", ""),
            )

        console.print(table)
    except Exception:
        print_error("Daemon not reachable.")
        raise typer.Exit(code=0)



"""RAG wrapper built-in tools: rag_query, rag_ingest."""

from __future__ import annotations

from astromesh.tools.base import BuiltinTool, ToolContext, ToolResult


class RagQueryTool(BuiltinTool):
    name = "rag_query"
    description = "Query the RAG pipeline for relevant documents"
    parameters = {
        "type": "object",
        "properties": {
            "query": {"type": "string"},
            "top_k": {"type": "integer", "default": 5},
        },
        "required": ["query"],
    }

    async def execute(self, arguments: dict, context: ToolContext) -> ToolResult:
        pipeline = getattr(context, "rag_pipeline", None)
        if pipeline is None:
            return ToolResult(
                success=False, data=None, error="RAG pipeline not available in context"
            )
        try:
            results = await pipeline.query(arguments["query"], top_k=arguments.get("top_k", 5))
            return ToolResult(success=True, data={"results": results}, metadata={})
        except Exception as e:
            return ToolResult(success=False, data=None, error=str(e))


class RagIngestTool(BuiltinTool):
    name = "rag_ingest"
    description = "Ingest a document into the RAG pipeline"
    parameters = {
        "type": "object",
        "properties": {
            "document": {"type": "string"},
            "metadata": {"type": "object"},
        },
        "required": ["document"],
    }

    async def execute(self, arguments: dict, context: ToolContext) -> ToolResult:
        pipeline = getattr(context, "rag_pipeline", None)
        if pipeline is None:
            return ToolResult(
                success=False, data=None, error="RAG pipeline not available in context"
            )
        try:
            await pipeline.ingest(arguments["document"], metadata=arguments.get("metadata", {}))
            return ToolResult(success=True, data={"ingested": True}, metadata={})
        except Exception as e:
            return ToolResult(success=False, data=None, error=str(e))



from __future__ import annotations

from astromesh.rag.reranking.base import Reranker


class CrossEncoderReranker(Reranker):
    """Reranker using a cross-encoder model from sentence-transformers."""

    def __init__(self, model_name: str = "cross-encoder/ms-marco-MiniLM-L-6-v2"):
        self.model_name = model_name
        self._model = None

    def _get_model(self):
        if self._model is None:
            from sentence_transformers import CrossEncoder

            self._model = CrossEncoder(self.model_name)
        return self._model

    async def rerank(self, query: str, documents: list[dict], top_k: int = 5) -> list[dict]:
        if not documents:
            return []

        model = self._get_model()
        pairs = [[query, doc.get("content", "")] for doc in documents]
        scores = model.predict(pairs)

        scored_docs = []
        for doc, score in zip(documents, scores):
            scored_docs.append({**doc, "rerank_score": float(score)})

        scored_docs.sort(key=lambda d: d["rerank_score"], reverse=True)
        return scored_docs[:top_k]



"""Tests for native routing helpers."""
import pytest


class TestRoutingHelpers:
    def test_ema_update(self):
        try:
            from astromesh._native import rust_ema_update
            result = rust_ema_update(100.0, 50.0, 0.8, 0.2)
            assert abs(result - 90.0) < 0.001
        except ImportError:
            pytest.skip("Native module not available")

    def test_detect_vision_no_images(self):
        try:
            from astromesh._native import rust_detect_vision
            messages = [{"role": "user", "content": "Hello"}]
            assert rust_detect_vision(messages) is False
        except ImportError:
            pytest.skip("Native module not available")

    def test_detect_vision_with_images(self):
        try:
            from astromesh._native import rust_detect_vision
            messages = [{"role": "user", "content": [{"type": "image_url", "image_url": {"url": "data:..."}}]}]
            assert rust_detect_vision(messages) is True
        except ImportError:
            pytest.skip("Native module not available")

    def test_rank_cost_optimized(self):
        try:
            from astromesh._native import rust_rank_candidates
            providers = [
                ("expensive", 0.05, 100.0, False, 0.0, True, True),
                ("cheap", 0.01, 200.0, False, 0.0, True, True),
                ("mid", 0.03, 150.0, False, 0.0, True, True),
            ]
            result = rust_rank_candidates(providers, "cost_optimized", 0, False, False)
            assert result[0] == "cheap"
            assert result[-1] == "expensive"
        except ImportError:
            pytest.skip("Native module not available")

    def test_rank_latency_optimized(self):
        try:
            from astromesh._native import rust_rank_candidates
            providers = [
                ("slow", 0.01, 300.0, False, 0.0, True, True),
                ("fast", 0.05, 50.0, False, 0.0, True, True),
            ]
            result = rust_rank_candidates(providers, "latency_optimized", 0, False, False)
            assert result[0] == "fast"
        except ImportError:
            pytest.skip("Native module not available")



"""Auto-detection of the platform's installer."""

from __future__ import annotations

import sys

from astromesh_node.installer.base import InstallerProtocol


def get_installer() -> InstallerProtocol:
    """Return the appropriate Installer for the current platform."""
    if sys.platform == "linux":
        from astromesh_node.installer.linux import LinuxInstaller

        return LinuxInstaller()
    elif sys.platform == "darwin":
        from astromesh_node.installer.macos import MacOSInstaller

        return MacOSInstaller()
    elif sys.platform == "win32":
        from astromesh_node.installer.windows import WindowsInstaller

        return WindowsInstaller()
    else:
        from astromesh_node.platform.base import UnsupportedPlatformError

        raise UnsupportedPlatformError(f"Platform {sys.platform} is not supported")



import os
import re

from astromesh.rag.chunking.base import ChunkingStrategy

try:
    from astromesh._native import rust_sentence_chunk as _native_chunk
except ImportError:
    _native_chunk = None


class SentenceChunker(ChunkingStrategy):
    def __init__(self, chunk_size: int = 500, overlap: int = 0):
        self.chunk_size = chunk_size
        self.overlap = overlap

    def chunk(self, document: str, metadata: dict) -> list[dict]:
        if not document:
            return []

        if _native_chunk is not None and not os.environ.get("ASTROMESH_FORCE_PYTHON"):
            return _native_chunk(document, metadata, self.chunk_size, self.overlap)

        sentences = self._split_sentences(document)
        chunks: list[dict] = []
        current = ""
        i = 0

        for sentence in sentences:
            if current and len(current) + len(sentence) > self.chunk_size:
                chunks.append(
                    {
                        "content": current.strip(),
                        "metadata": {**metadata, "chunk_index": i, "strategy": "sentence"},
                    }
                )
                i += 1
                current = sentence
            else:
                current += sentence

        if current.strip():
            chunks.append(
                {
                    "content": current.strip(),
                    "metadata": {**metadata, "chunk_index": i, "strategy": "sentence"},
                }
            )

        return chunks

    @staticmethod
    def _split_sentences(text: str) -> list[str]:
        """Split text on sentence boundaries, keeping delimiters attached."""
        parts = re.split(r"(?<=[.!?])\s+", text)
        # Re-add trailing space for all but the last
        sentences: list[str] = []
        for idx, part in enumerate(parts):
            if idx < len(parts) - 1:
                sentences.append(part + " ")
            else:
                sentences.append(part)
        return sentences



import json

import asyncpg

from astromesh.core.memory import ConversationBackend, ConversationTurn


class PGConversationBackend(ConversationBackend):
    def __init__(self, pool: asyncpg.Pool | None = None, dsn: str | None = None):
        self._pool = pool
        self._dsn = dsn

    async def initialize(self):
        if self._pool is None and self._dsn:
            self._pool = await asyncpg.create_pool(self._dsn)
        async with self._pool.acquire() as conn:
            await conn.execute(
                "CREATE TABLE IF NOT EXISTS conversation_history ("
                "id SERIAL PRIMARY KEY, "
                "session_id TEXT, "
                "role TEXT, "
                "content TEXT, "
                "metadata JSONB DEFAULT '{}', "
                "token_count INTEGER DEFAULT 0, "
                "timestamp TIMESTAMPTZ"
                ")"
            )
            await conn.execute(
                "CREATE TABLE IF NOT EXISTS conversation_summaries ("
                "session_id TEXT PRIMARY KEY, "
                "summary TEXT"
                ")"
            )

    async def save_turn(self, session_id, turn):
        async with self._pool.acquire() as conn:
            await conn.execute(
                "INSERT INTO conversation_history "
                "(session_id, role, content, metadata, token_count, timestamp) "
                "VALUES ($1, $2, $3, $4, $5, $6)",
                session_id,
                turn.role,
                turn.content,
                json.dumps(turn.metadata),
                turn.token_count,
                turn.timestamp,
            )

    async def get_history(self, session_id, limit=50):
        async with self._pool.acquire() as conn:
            rows = await conn.fetch(
                "SELECT role, content, metadata, token_count, timestamp "
                "FROM conversation_history WHERE session_id = $1 "
                "ORDER BY id DESC LIMIT $2",
                session_id,
                limit,
            )
        rows = list(reversed(rows))
        return [
            ConversationTurn(
                role=row["role"],
                content=row["content"],
                timestamp=row["timestamp"],
                metadata=json.loads(row["metadata"])
                if isinstance(row["metadata"], str)
                else row["metadata"],
                token_count=row["token_count"] or 0,
            )
            for row in rows
        ]

    async def clear(self, session_id):
        async with self._pool.acquire() as conn:
            await conn.execute("DELETE FROM conversation_history WHERE session_id = $1", session_id)
            await conn.execute(
                "DELETE FROM conversation_summaries WHERE session_id = $1", session_id
            )

    async def get_summary(self, session_id):
        async with self._pool.acquire() as conn:
            row = await conn.fetchrow(
                "SELECT summary FROM conversation_summaries WHERE session_id = $1",
                session_id,
            )
        return row["summary"] if row else None

    async def save_summary(self, session_id, summary):
        async with self._pool.acquire() as conn:
            await conn.execute(
                "INSERT INTO conversation_summaries (session_id, summary) "
                "VALUES ($1, $2) "
                "ON CONFLICT (session_id) DO UPDATE SET summary = $2",
                session_id,
                summary,
            )



"""Built-in tool implementations for Astromesh."""

from astromesh.tools.builtin.utilities import (
    DatetimeNowTool,
    JsonTransformTool,
    CacheStoreTool,
)
from astromesh.tools.builtin.http import HttpRequestTool, GraphQLQueryTool
from astromesh.tools.builtin.web_search import WebSearchTool, WebScrapeTool, WikipediaTool
from astromesh.tools.builtin.files import ReadFileTool, WriteFileTool
from astromesh.tools.builtin.database import SqlQueryTool
from astromesh.tools.builtin.communication import (
    SendWebhookTool,
    SendSlackTool,
    SendEmailTool,
)
from astromesh.tools.builtin.ai import TextSummarizeTool
from astromesh.tools.builtin.rag import RagQueryTool, RagIngestTool

ALL_TOOLS: list = [
    DatetimeNowTool,
    JsonTransformTool,
    CacheStoreTool,
    HttpRequestTool,
    GraphQLQueryTool,
    WebSearchTool,
    WebScrapeTool,
    WikipediaTool,
    ReadFileTool,
    WriteFileTool,
    SqlQueryTool,
    SendWebhookTool,
    SendSlackTool,
    SendEmailTool,
    TextSummarizeTool,
    RagQueryTool,
    RagIngestTool,
]



"""Tests for astromesh.mesh.leader — bully algorithm leader election."""

import time

import pytest

from astromesh.mesh.config import MeshConfig
from astromesh.mesh.leader import LeaderElector
from astromesh.mesh.manager import MeshManager
from astromesh.mesh.state import NodeLoad, NodeState
from astromesh.runtime.services import ServiceManager


@pytest.fixture
def mesh():
    config = MeshConfig(enabled=True, node_name="test-node")
    sm = ServiceManager({"api": True, "agents": True})
    return MeshManager(config, sm)


@pytest.fixture
def elector(mesh):
    return LeaderElector(mesh)


def _make_node(
    node_id: str,
    name: str = "peer",
    status: str = "alive",
) -> NodeState:
    return NodeState(
        node_id=node_id,
        name=name,
        url="http://localhost:9000",
        services=["api"],
        agents=[],
        load=NodeLoad(),
        joined_at=time.time(),
        last_heartbeat=time.time(),
        status=status,
    )


def test_elector_init(elector):
    assert elector.current_leader() is None
    assert elector.is_leader() is False


def test_elector_elect_self_when_alone(mesh, elector):
    leader_id = elector.elect()
    assert leader_id == mesh.node_id
    assert elector.is_leader() is True
    assert elector.current_leader() == mesh.node_id


def test_elector_highest_id_wins(mesh, elector):
    high_node = _make_node(node_id="zzz-highest", name="high-peer")
    mesh.cluster_state().add_node(high_node)

    leader_id = elector.elect()
    assert leader_id == "zzz-highest"
    assert elector.is_leader() is False
    assert mesh.cluster_state().leader_id == "zzz-highest"
    assert mesh.cluster_state().nodes["zzz-highest"].leader is True


def test_elector_ignores_dead_nodes(mesh, elector):
    dead_node = _make_node(node_id="zzz-dead", name="dead-peer", status="dead")
    mesh.cluster_state().add_node(dead_node)

    leader_id = elector.elect()
    assert leader_id == mesh.node_id
    assert elector.is_leader() is True
    assert mesh.cluster_state().nodes["zzz-dead"].leader is False


def test_elector_on_node_failed_triggers_election(mesh, elector):
    high_node = _make_node(node_id="zzz-highest", name="high-peer")
    mesh.cluster_state().add_node(high_node)
    elector.elect()
    assert elector.current_leader() == "zzz-highest"

    # Simulate failure: mark node as dead
    mesh.cluster_state().nodes["zzz-highest"].status = "dead"
    elector.on_node_failed("zzz-highest")

    # New leader should be the local node (only alive node left)
    assert elector.current_leader() == mesh.node_id
    assert elector.is_leader() is True


def test_elector_on_node_joined_higher(mesh, elector):
    elector.elect()
    assert elector.is_leader() is True

    # A node with higher ID joins
    high_node = _make_node(node_id="zzz-highest", name="high-peer")
    mesh.cluster_state().add_node(high_node)
    elector.on_node_joined("zzz-highest")

    assert elector.current_leader() == "zzz-highest"
    assert elector.is_leader() is False



"""RunResult and StreamEvent types."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any


@dataclass
class RunResult:
    """Result of an agent execution."""

    answer: str
    steps: list[dict]
    trace: dict | None
    cost: float
    tokens: dict[str, int]
    latency_ms: float
    model: str
    metadata: dict = field(default_factory=dict)

    @classmethod
    def from_runtime(cls, runtime_result: dict) -> RunResult:
        """Build RunResult from the dict returned by AgentRuntime.run()."""
        trace = runtime_result.get("trace")
        cost = 0.0
        input_tokens = 0
        output_tokens = 0
        latency_ms = 0.0
        model = ""

        if trace and isinstance(trace, dict):
            spans = trace.get("spans", [])
            for span in spans:
                attrs = span.get("attributes", {})
                if span.get("name") == "llm.complete":
                    cost += attrs.get("cost", 0.0)
                    input_tokens += attrs.get("input_tokens", 0)
                    output_tokens += attrs.get("output_tokens", 0)
                    model = attrs.get("model", model)
                if span.get("parent_span_id") is None:
                    latency_ms = span.get("duration_ms", 0.0)

        return cls(
            answer=runtime_result.get("answer", ""),
            steps=runtime_result.get("steps", []),
            trace=trace,
            cost=cost,
            tokens={"input": input_tokens, "output": output_tokens},
            latency_ms=latency_ms,
            model=model,
        )


@dataclass
class StreamEvent:
    """Event emitted during agent streaming."""

    type: str  # "step" | "token" | "done"
    step: dict | None = None
    content: str | None = None
    result: RunResult | None = None



import pytest
from asgi_lifespan import LifespanManager
from httpx import ASGITransport, AsyncClient

from astromesh.api.main import app
from astromesh.observability.collector import InternalCollector
from astromesh.observability.tracing import TracingContext
from astromesh.api.routes.traces import set_collector


@pytest.fixture
async def collector():
    c = InternalCollector()
    set_collector(c)
    ctx = TracingContext(agent_name="test-agent", session_id="s1")
    span = ctx.start_span("agent.run")
    span.set_attribute("tokens", 100)
    span.finish()
    await c.emit_trace(ctx)
    return c, ctx.trace_id


class TestTracesAPI:
    async def test_list_traces(self, collector):
        c, trace_id = collector
        async with LifespanManager(app):
            async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as client:
                resp = await client.get("/v1/traces/", params={"agent": "test-agent"})
        assert resp.status_code == 200
        data = resp.json()
        assert len(data["traces"]) == 1
        assert data["traces"][0]["trace_id"] == trace_id

    async def test_get_trace(self, collector):
        c, trace_id = collector
        async with LifespanManager(app):
            async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as client:
                resp = await client.get(f"/v1/traces/{trace_id}")
        assert resp.status_code == 200
        data = resp.json()
        assert data["trace_id"] == trace_id

    async def test_get_trace_not_found(self, collector):
        async with LifespanManager(app):
            async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as client:
                resp = await client.get("/v1/traces/nonexistent")
        assert resp.status_code == 404



"""WhatsApp channel webhook handler."""

from __future__ import annotations

import logging

from fastapi import APIRouter, BackgroundTasks, Query, Request, Response

from astromesh.channels.whatsapp import WhatsAppClient
from astromesh.channels.media import build_multimodal_query

logger = logging.getLogger(__name__)
router = APIRouter(tags=["channels"])

_whatsapp = WhatsAppClient()
_default_agent = "whatsapp-assistant"  # TODO: make configurable

_runtime = None


def set_runtime(runtime):
    global _runtime
    _runtime = runtime


@router.get("/channels/whatsapp/webhook")
async def verify_webhook(
    hub_mode: str = Query(None, alias="hub.mode"),
    hub_token: str = Query(None, alias="hub.verify_token"),
    hub_challenge: str = Query(None, alias="hub.challenge"),
):
    """Meta webhook verification endpoint."""
    result = _whatsapp.verify_webhook(hub_mode or "", hub_token or "", hub_challenge or "")
    if result is not None:
        return Response(content=result, media_type="text/plain")
    return Response(status_code=403)


async def _process_message(message):
    """Process incoming ChannelMessage in background."""
    try:
        # Download media and build the query (str or multimodal list).
        query = await build_multimodal_query(message, _whatsapp)
        result = await _runtime.run(
            agent_name=_default_agent,
            query=query,
            session_id=f"wa_{message.sender_id}",
        )
        answer = result.get("answer", "Lo siento, no pude procesar tu mensaje.")
        await _whatsapp.send_text(message.sender_id, answer)
    except Exception:
        logger.exception("Failed to process WhatsApp message from %s", message.sender_id)
        try:
            await _whatsapp.send_text(
                message.sender_id, "Lo siento, ocurrio un error. Intenta de nuevo."
            )
        except Exception:
            logger.exception("Failed to send error message to %s", message.sender_id)


@router.post("/channels/whatsapp/webhook")
async def receive_message(request: Request, background_tasks: BackgroundTasks):
    """Receive incoming WhatsApp messages from Meta webhook."""
    body = await request.body()
    signature = request.headers.get("X-Hub-Signature-256", "")

    if not _whatsapp.verify_request(body, signature):
        return Response(status_code=403)

    payload = await request.json()
    messages = await _whatsapp.parse_incoming(payload)

    for msg in messages:
        logger.info("WhatsApp message from %s: %s", msg.sender_id, msg.message_id)
        background_tasks.add_task(_process_message, msg)

    return {"status": "ok"}



"""HuggingFace Text Generation Inference provider adapter for the Astromesh Agent Runtime."""

from __future__ import annotations

import time
from typing import Any, AsyncIterator

import httpx

from .base import CompletionChunk, CompletionResponse


class HFTGIProvider:
    """Provider adapter for HuggingFace TGI (OpenAI-compatible)."""

    def __init__(self, config: dict[str, Any] | None = None) -> None:
        config = config or {}
        self.base_url: str = config.get("base_url", "http://localhost:8080")
        self.model: str = config.get("model", "default")
        self.timeout: float = config.get("timeout", 120.0)
        self._client: httpx.AsyncClient | None = None

    async def _get_client(self) -> httpx.AsyncClient:
        if self._client is None or self._client.is_closed:
            self._client = httpx.AsyncClient(
                base_url=self.base_url,
                timeout=self.timeout,
            )
        return self._client

    async def complete(self, messages: list[dict], **kwargs: Any) -> CompletionResponse:
        client = await self._get_client()
        model = kwargs.pop("model", self.model)

        payload: dict[str, Any] = {
            "model": model,
            "messages": messages,
            "stream": False,
        }
        payload.update(kwargs)

        start = time.perf_counter()
        resp = await client.post("/v1/chat/completions", json=payload)
        latency_ms = (time.perf_counter() - start) * 1000
        resp.raise_for_status()
        data = resp.json()

        choice = data["choices"][0]
        message = choice.get("message", {})
        usage_data = data.get("usage", {})
        input_tokens = usage_data.get("prompt_tokens", 0)
        output_tokens = usage_data.get("completion_tokens", 0)

        return CompletionResponse(
            content=message.get("content", "") or "",
            model=model,
            provider="hf_tgi",
            usage={"input_tokens": input_tokens, "output_tokens": output_tokens},
            latency_ms=latency_ms,
            cost=0.0,
            tool_calls=message.get("tool_calls", []),
        )

    async def stream(self, messages: list[dict], **kwargs: Any) -> AsyncIterator[CompletionChunk]:
        client = await self._get_client()
        model = kwargs.pop("model", self.model)

        payload: dict[str, Any] = {
            "model": model,
            "messages": messages,
            "stream": True,
        }
        payload.update(kwargs)

        async with client.stream("POST", "/v1/chat/completions", json=payload) as resp:
            resp.raise_for_status()
            async for line in resp.aiter_lines():
                if not line or not line.startswith("data: "):
                    continue
                raw = line[len("data: ") :]
                if raw.strip() == "[DONE]":
                    yield CompletionChunk(content="", model=model, provider="hf_tgi", done=True)
                    break
                import json

                chunk_data = json.loads(raw)
                delta = chunk_data["choices"][0].get("delta", {})
                yield CompletionChunk(
                    content=delta.get("content", "") or "",
                    model=model,
                    provider="hf_tgi",
                    done=False,
                )

    async def health_check(self) -> bool:
        try:
            client = await self._get_client()
            resp = await client.get("/health")
            return resp.status_code == 200
        except Exception:
            return False

    def supports_tools(self) -> bool:
        return True

    def supports_vision(self) -> bool:
        return False

    def estimated_cost(self, model: str, input_tokens: int, output_tokens: int) -> float:
        return 0.0



"""Tests for MeshConfig."""

from astromesh.mesh.config import MeshConfig


def test_mesh_config_defaults():
    cfg = MeshConfig()
    assert cfg.enabled is False
    assert cfg.node_name == ""
    assert cfg.bind == "0.0.0.0:8000"
    assert cfg.seeds == []
    assert cfg.heartbeat_interval == 5
    assert cfg.gossip_interval == 2
    assert cfg.gossip_fanout == 3
    assert cfg.failure_timeout == 15
    assert cfg.dead_timeout == 30


def test_mesh_config_custom():
    cfg = MeshConfig(
        enabled=True,
        node_name="node-1",
        bind="127.0.0.1:9000",
        seeds=["10.0.0.1:8000", "10.0.0.2:8000"],
        heartbeat_interval=10,
        gossip_interval=4,
        gossip_fanout=5,
        failure_timeout=30,
        dead_timeout=60,
    )
    assert cfg.enabled is True
    assert cfg.node_name == "node-1"
    assert cfg.bind == "127.0.0.1:9000"
    assert cfg.seeds == ["10.0.0.1:8000", "10.0.0.2:8000"]
    assert cfg.heartbeat_interval == 10
    assert cfg.gossip_interval == 4
    assert cfg.gossip_fanout == 5
    assert cfg.failure_timeout == 30
    assert cfg.dead_timeout == 60


def test_mesh_config_from_dict():
    data = {"enabled": True, "node_name": "alpha", "seeds": ["peer:8000"]}
    cfg = MeshConfig.from_dict(data)
    assert cfg.enabled is True
    assert cfg.node_name == "alpha"
    assert cfg.seeds == ["peer:8000"]
    # Defaults for missing keys
    assert cfg.bind == "0.0.0.0:8000"
    assert cfg.heartbeat_interval == 5
    assert cfg.gossip_interval == 2
    assert cfg.gossip_fanout == 3
    assert cfg.failure_timeout == 15
    assert cfg.dead_timeout == 30


def test_mesh_config_from_dict_empty():
    cfg = MeshConfig.from_dict({})
    assert cfg.enabled is False
    assert cfg.node_name == ""
    assert cfg.bind == "0.0.0.0:8000"
    assert cfg.seeds == []
    assert cfg.heartbeat_interval == 5


def test_mesh_config_from_dict_none():
    cfg = MeshConfig.from_dict(None)
    assert cfg.enabled is False
    assert cfg.node_name == ""
    assert cfg.bind == "0.0.0.0:8000"
    assert cfg.seeds == []
    assert cfg.heartbeat_interval == 5



# astromesh/api/routes/workflows.py
from __future__ import annotations

from typing import Any

from fastapi import APIRouter, HTTPException
from pydantic import BaseModel

router = APIRouter(prefix="/workflows", tags=["workflows"])

_engine = None


def set_workflow_engine(engine) -> None:
    global _engine
    _engine = engine


class WorkflowRunRequest(BaseModel):
    query: str = ""
    trigger: dict[str, Any] | None = None


@router.get("/")
async def list_workflows():
    if not _engine:
        return {"workflows": []}
    return {"workflows": _engine.list_workflows()}


@router.get("/{name}")
async def get_workflow(name: str):
    if not _engine:
        raise HTTPException(status_code=503, detail="Workflow engine not initialized")
    wf = _engine.get_workflow(name)
    if not wf:
        raise HTTPException(status_code=404, detail=f"Workflow '{name}' not found")
    return {
        "name": wf.name,
        "version": wf.version,
        "namespace": wf.namespace,
        "description": wf.description,
        "trigger": wf.trigger,
        "timeout_seconds": wf.timeout_seconds,
        "steps": [
            {"name": s.name, "type": s.step_type.value}
            for s in wf.steps
        ],
    }


@router.post("/{name}/run")
async def run_workflow(name: str, request: WorkflowRunRequest):
    if not _engine:
        raise HTTPException(status_code=503, detail="Workflow engine not initialized")
    trigger = request.trigger or {"query": request.query}
    try:
        result = await _engine.run(name, trigger=trigger)
    except ValueError as e:
        raise HTTPException(status_code=404, detail=str(e))
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))
    return {
        "workflow_name": result.workflow_name,
        "status": result.status,
        "output": result.output,
        "steps": {
            k: {"status": v.status.value, "output": v.output, "error": v.error}
            for k, v in result.steps.items()
        },
        "trace": result.trace,
        "duration_ms": result.duration_ms,
    }



"""Tests for all 6 Astromesh provider adapters."""

from __future__ import annotations

import httpx
import respx

from astromesh.providers.base import CompletionResponse
from astromesh.providers.hf_tgi_provider import HFTGIProvider
from astromesh.providers.llamacpp_provider import LlamaCppProvider
from astromesh.providers.ollama_provider import OllamaProvider
from astromesh.providers.onnx_provider import ONNXProvider
from astromesh.providers.openai_compat import OpenAICompatProvider
from astromesh.providers.vllm_provider import VLLMProvider

# ---------------------------------------------------------------------------
# Helpers — canonical mock responses
# ---------------------------------------------------------------------------

OLLAMA_CHAT_RESPONSE = {
    "message": {"role": "assistant", "content": "Hello from Ollama!"},
    "prompt_eval_count": 10,
    "eval_count": 20,
}

OPENAI_CHAT_RESPONSE = {
    "id": "chatcmpl-test",
    "object": "chat.completion",
    "choices": [
        {
            "index": 0,
            "message": {"role": "assistant", "content": "Hello from OpenAI!"},
            "finish_reason": "stop",
        }
    ],
    "usage": {"prompt_tokens": 15, "completion_tokens": 25, "total_tokens": 40},
}

MESSAGES = [{"role": "user", "content": "Hi"}]


# ===================================================================
# Ollama
# ===================================================================


@respx.mock
async def test_ollama_complete():
    respx.post("http://localhost:11434/api/chat").mock(
        return_value=httpx.Response(200, json=OLLAMA_CHAT_RESPONSE)
    )

    provider = OllamaProvider({"base_url": "http://localhost:11434", "model": "llama3"})
    result = await provider.complete(MESSAGES)

    assert isinstance(result, CompletionResponse)
    assert result.content == "Hello from Ollama!"
    assert result.provider == "ollama"
    assert result.usage["input_tokens"] == 10
    assert result.usage["output_tokens"] == 20
    assert result.cost == 0.0


@respx.mock
async def test_ollama_health_check_healthy():
    respx.get("http://localhost:11434/api/tags").mock(
        return_value=httpx.Response(200, json={"models": []})
    )

    provider = OllamaProvider({"base_url": "http://localhost:11434"})
    assert await provider.health_check() is True


@respx.mock
async def test_ollama_health_check_unhealthy():
    respx.get("http://localhost:11434/api/tags").mock(
        return_value=httpx.Response(500)
    )

    provider = OllamaProvider({"base_url": "http://localhost:11434"})
    assert await provider.health_check() is False


def test_ollama_supports_tools():
    provider = OllamaProvider({"model": "llama3"})
    assert provider.supports_tools() is True


def test_ollama_supports_vision():
    assert OllamaProvider({"model": "llava"}).supports_vision() is True
    assert OllamaProvider({"model": "llava-13b"}).supports_vision() is True
    assert OllamaProvider({"model": "my-vision-model"}).supports_vision() is True
    assert OllamaProvider({"model": "llama3"}).supports_vision() is False


# ===================================================================
# OpenAI-Compatible
# ===================================================================


@respx.mock
async def test_openai_compat_complete():
    respx.post("https://api.openai.com/v1/chat/completions").mock(
        return_value=httpx.Response(200, json=OPENAI_CHAT_RESPONSE)
    )

    provider = OpenAICompatProvider(
        {"base_url": "https://api.openai.com/v1", "model": "gpt-4o", "api_key": "sk-test"}
    )
    result = await provider.complete(MESSAGES)

    assert isinstance(result, CompletionResponse)
    assert result.content == "Hello from OpenAI!"
    assert result.provider == "openai_compat"
    assert result.usage["input_tokens"] == 15
    assert result.usage["output_tokens"] == 25
    assert result.cost > 0


def test_openai_compat_supports_tools():
    provider = OpenAICompatProvider({"model": "gpt-4o", "api_key": "sk-test"})
    assert provider.supports_tools() is True


# ===================================================================
# vLLM
# ===================================================================


@respx.mock
async def test_vllm_complete():
    respx.post("http://localhost:8000/v1/chat/completions").mock(
        return_value=httpx.Response(200, json=OPENAI_CHAT_RESPONSE)
    )

    provider = VLLMProvider({"base_url": "http://localhost:8000", "model": "my-model"})
    result = await provider.complete(MESSAGES)

    assert isinstance(result, CompletionResponse)
    assert result.content == "Hello from OpenAI!"
    assert result.provider == "vllm"
    assert result.usage["input_tokens"] == 15


# ===================================================================
# llama.cpp
# ===================================================================


@respx.mock
async def test_llamacpp_complete():
    respx.post("http://localhost:8080/v1/chat/completions").mock(
        return_value=httpx.Response(200, json=OPENAI_CHAT_RESPONSE)
    )

    provider = LlamaCppProvider({"base_url": "http://localhost:8080", "model": "local-llama"})
    result = await provider.complete(MESSAGES)

    assert isinstance(result, CompletionResponse)
    assert result.content == "Hello from OpenAI!"
    assert result.provider == "llamacpp"
    assert result.cost == 0.0


# ===================================================================
# HuggingFace TGI
# ===================================================================


@respx.mock
async def test_hf_tgi_complete():
    respx.post("http://localhost:8080/v1/chat/completions").mock(
        return_value=httpx.Response(200, json=OPENAI_CHAT_RESPONSE)
    )

    provider = HFTGIProvider({"base_url": "http://localhost:8080", "model": "tgi-model"})
    result = await provider.complete(MESSAGES)

    assert isinstance(result, CompletionResponse)
    assert result.content == "Hello from OpenAI!"
    assert result.provider == "hf_tgi"
    assert result.cost == 0.0


# ===================================================================
# ONNX
# ===================================================================


def test_onnx_provider_init():
    provider = ONNXProvider({"model_path": "/tmp/model.onnx", "model": "my-onnx"})
    assert provider.model_path == "/tmp/model.onnx"
    assert provider.model_name == "my-onnx"
    assert provider.supports_tools() is False
    assert provider.supports_vision() is False
    assert provider.estimated_cost("any", 100, 100) == 0.0
    assert provider._session is None



"""Tests for ServiceManager."""

from astromesh.runtime.services import DEFAULT_SERVICES, ServiceManager


def test_default_services():
    sm = ServiceManager({})
    for service in DEFAULT_SERVICES:
        assert sm.is_enabled(service) is True


def test_explicit_disable():
    sm = ServiceManager({"inference": False, "channels": False})
    assert sm.is_enabled("inference") is False
    assert sm.is_enabled("channels") is False
    assert sm.is_enabled("agents") is True
    assert sm.is_enabled("api") is True


def test_enabled_services_list():
    sm = ServiceManager({"inference": False, "rag": False})
    enabled = sm.enabled_services()
    assert "inference" not in enabled
    assert "rag" not in enabled
    assert "agents" in enabled
    assert "api" in enabled


def test_unknown_service():
    sm = ServiceManager({})
    assert sm.is_enabled("nonexistent") is False


def test_validate_warns_agents_without_tools():
    sm = ServiceManager({"agents": True, "tools": False})
    warnings = sm.validate()
    assert any("tools" in w.lower() for w in warnings)


def test_validate_no_warnings_when_consistent():
    sm = ServiceManager({"agents": True, "tools": True})
    warnings = sm.validate()
    assert len(warnings) == 0


def test_to_dict():
    sm = ServiceManager({"inference": False})
    d = sm.to_dict()
    assert d["inference"] is False
    assert d["api"] is True
    assert isinstance(d, dict)



"""Tests that Rust and Python guardrails produce identical results."""
import pytest


class TestPiiRedaction:
    def test_email_redaction(self, use_native):
        from astromesh.core.guardrails import GuardrailsEngine
        engine = GuardrailsEngine({"input": [{"type": "pii_detection", "action": "redact"}]})
        result = engine._redact_pii("Contact john@example.com for info", "redact")
        assert "[REDACTED_EMAIL]" in result
        assert "john@example.com" not in result

    def test_phone_redaction(self, use_native):
        from astromesh.core.guardrails import GuardrailsEngine
        engine = GuardrailsEngine()
        result = engine._redact_pii("Call 555-123-4567 now", "redact")
        assert "[REDACTED_PHONE]" in result

    def test_ssn_redaction(self, use_native):
        from astromesh.core.guardrails import GuardrailsEngine
        engine = GuardrailsEngine()
        result = engine._redact_pii("SSN: 123-45-6789", "redact")
        assert "[REDACTED_SSN]" in result

    def test_cc_redaction(self, use_native):
        from astromesh.core.guardrails import GuardrailsEngine
        engine = GuardrailsEngine()
        result = engine._redact_pii("Card: 4111 1111 1111 1111", "redact")
        assert "[REDACTED_CC]" in result

    def test_no_pii(self, use_native):
        from astromesh.core.guardrails import GuardrailsEngine
        engine = GuardrailsEngine()
        text = "This is a normal text with no PII."
        result = engine._redact_pii(text, "redact")
        assert result == text

    def test_action_not_redact(self, use_native):
        from astromesh.core.guardrails import GuardrailsEngine
        engine = GuardrailsEngine()
        text = "Contact john@example.com"
        result = engine._redact_pii(text, "warn")
        assert result == text


class TestTopicFilter:
    @pytest.mark.asyncio
    async def test_blocked_topic_raises(self, use_native):
        from astromesh.core.guardrails import GuardrailsEngine
        engine = GuardrailsEngine({
            "input": [{"type": "topic_filter", "blocked_topics": ["violence"], "action": "block"}]
        })
        with pytest.raises(ValueError, match="Blocked topic detected"):
            await engine.apply_input("Let's discuss violence in media")

    @pytest.mark.asyncio
    async def test_no_blocked_topic(self, use_native):
        from astromesh.core.guardrails import GuardrailsEngine
        engine = GuardrailsEngine({
            "input": [{"type": "topic_filter", "blocked_topics": ["violence"], "action": "block"}]
        })
        result = await engine.apply_input("Let's discuss cooking recipes")
        assert result == "Let's discuss cooking recipes"



"""Installer abstraction — platform-specific filesystem and service setup."""

from astromesh_node.installer.base import InstallerProtocol
from astromesh_node.installer.detect import get_installer

__all__ = ["InstallerProtocol", "get_installer"]



"""astromeshctl run — Execute agents and workflows."""

import json
import uuid
from typing import Optional

import typer
from rich.panel import Panel
from rich.table import Table

from astromesh_cli.client import api_post_with_timeout
from astromesh_cli.output import console, print_error, print_json


def run_command(
    name: str = typer.Argument(..., help="Agent or workflow name to run"),
    query: str = typer.Argument("", help="Query to send to the agent"),
    session: Optional[str] = typer.Option(
        None, "--session", help="Session ID (auto-generated if not set)"
    ),
    json_output: bool = typer.Option(False, "--json", help="Output raw JSON response"),
    timeout: float = typer.Option(60.0, "--timeout", help="Request timeout in seconds"),
    workflow: bool = typer.Option(False, "--workflow", help="Run as workflow instead of agent"),
    input_data: Optional[str] = typer.Option(None, "--input", help="Workflow input data (JSON)"),
) -> None:
    """Execute an agent with a query or run a workflow."""
    if workflow:
        _run_workflow(name, query, input_data, json_output, timeout)
        return

    if not query:
        print_error("Query is required when running an agent.")
        raise typer.Exit(code=1)

    session_id = session or str(uuid.uuid4())

    try:
        data = api_post_with_timeout(
            f"/v1/agents/{name}/run",
            json={"query": query, "session_id": session_id},
            timeout=timeout,
        )
    except Exception as e:
        print_error(f"Failed to run agent '{name}': {e}")
        raise typer.Exit(code=1)

    if json_output:
        print_json(data)
        return

    response_text = data.get("response", "")
    trace_id = data.get("trace_id", "N/A")
    tokens = data.get("tokens_used", "N/A")

    console.print(
        Panel(
            response_text,
            title=f"[cyan]{name}[/cyan]",
            subtitle=f"trace: {trace_id} | tokens: {tokens}",
            border_style="blue",
        )
    )


def _run_workflow(
    name: str, query: str, input_data: str | None, json_output: bool, timeout: float
) -> None:
    """Execute a workflow via the API."""
    if input_data:
        try:
            trigger = json.loads(input_data)
        except json.JSONDecodeError as e:
            print_error(f"Invalid JSON in --input: {e}")
            raise typer.Exit(code=1)
        payload = {"trigger": trigger, "query": query}
    else:
        payload = {"query": query}

    try:
        data = api_post_with_timeout(
            f"/v1/workflows/{name}/run",
            json=payload,
            timeout=timeout,
        )
    except Exception as e:
        print_error(f"Failed to run workflow '{name}': {e}")
        raise typer.Exit(code=1)

    if json_output:
        print_json(data)
        return

    status = data.get("status", "unknown")
    duration = data.get("duration_ms", 0)
    output = data.get("output", {})
    steps = data.get("steps", {})

    # Build step summary table
    table = Table(title="Steps", show_header=True)
    table.add_column("Step", style="cyan")
    table.add_column("Status")
    for step_name, step_info in steps.items():
        step_status = step_info.get("status", "unknown")
        style = "green" if step_status == "success" else "red"
        table.add_row(step_name, f"[{style}]{step_status}[/{style}]")

    answer = output.get("answer", str(output)) if isinstance(output, dict) else str(output)
    console.print(
        Panel(
            answer,
            title=f"[cyan]workflow:{name}[/cyan]",
            subtitle=f"status: {status} | {duration:.0f}ms",
            border_style="green" if status == "completed" else "red",
        )
    )
    if steps:
        console.print(table)



from astromesh.core.prompt_engine import PromptEngine


def test_render_simple_template():
    engine = PromptEngine()
    result = engine.render("Hello, {{ name }}!", {"name": "World"})
    assert result == "Hello, World!"


def test_render_with_missing_var():
    engine = PromptEngine()
    result = engine.render("Hello, {{ name }}!", {})
    assert result == "Hello, !"


def test_render_multiline():
    engine = PromptEngine()
    template = """You are {{ role }}.
Your task is {{ task }}.
{% for item in items %}
- {{ item }}
{% endfor %}"""
    result = engine.render(template, {
        "role": "an assistant",
        "task": "helping users",
        "items": ["search", "answer", "summarize"],
    })
    assert "an assistant" in result
    assert "helping users" in result
    assert "- search" in result
    assert "- answer" in result
    assert "- summarize" in result


def test_register_and_render_template():
    engine = PromptEngine()
    engine.register_template("greeting", "Hi {{ user }}, welcome!")
    result = engine.render_template("greeting", {"user": "Alice"})
    assert result == "Hi Alice, welcome!"


def test_render_nonexistent_template():
    engine = PromptEngine()
    result = engine.render_template("does_not_exist", {"x": 1})
    assert result == ""



from dataclasses import dataclass


@dataclass
class MetricsConfig:
    enabled: bool = True
    prefix: str = "astromesh"


class MetricsCollector:
    """Prometheus metrics for Astromesh agent runs."""

    def __init__(self, config: MetricsConfig | None = None):
        self._config = config or MetricsConfig()
        self._counters = {}
        self._histograms = {}
        self._gauges = {}

        if self._config.enabled:
            self._setup_metrics()

    def _setup_metrics(self):
        try:
            from prometheus_client import Counter, Histogram, Gauge

            prefix = self._config.prefix
            self._counters["agent_runs"] = Counter(
                f"{prefix}_agent_runs_total",
                "Total agent runs",
                ["agent_name", "pattern", "status"],
            )
            self._counters["provider_calls"] = Counter(
                f"{prefix}_provider_calls_total",
                "Total provider calls",
                ["provider", "model", "status"],
            )
            self._counters["tool_executions"] = Counter(
                f"{prefix}_tool_executions_total",
                "Total tool executions",
                ["tool_name", "status"],
            )
            self._counters["tokens_used"] = Counter(
                f"{prefix}_tokens_used_total",
                "Total tokens used",
                ["agent_name", "direction"],  # direction: input/output
            )
            self._histograms["agent_latency"] = Histogram(
                f"{prefix}_agent_latency_seconds",
                "Agent run latency",
                ["agent_name", "pattern"],
            )
            self._histograms["provider_latency"] = Histogram(
                f"{prefix}_provider_latency_seconds",
                "Provider call latency",
                ["provider", "model"],
            )
            self._gauges["active_sessions"] = Gauge(
                f"{prefix}_active_sessions",
                "Currently active sessions",
                ["agent_name"],
            )
        except ImportError:
            pass

    def record_agent_run(self, agent_name: str, pattern: str, status: str, latency_s: float):
        if "agent_runs" in self._counters:
            self._counters["agent_runs"].labels(
                agent_name=agent_name, pattern=pattern, status=status
            ).inc()
        if "agent_latency" in self._histograms:
            self._histograms["agent_latency"].labels(
                agent_name=agent_name, pattern=pattern
            ).observe(latency_s)

    def record_provider_call(self, provider: str, model: str, status: str, latency_s: float):
        if "provider_calls" in self._counters:
            self._counters["provider_calls"].labels(
                provider=provider, model=model, status=status
            ).inc()
        if "provider_latency" in self._histograms:
            self._histograms["provider_latency"].labels(provider=provider, model=model).observe(
                latency_s
            )

    def record_tool_execution(self, tool_name: str, status: str):
        if "tool_executions" in self._counters:
            self._counters["tool_executions"].labels(tool_name=tool_name, status=status).inc()

    def record_tokens(self, agent_name: str, input_tokens: int, output_tokens: int):
        if "tokens_used" in self._counters:
            self._counters["tokens_used"].labels(agent_name=agent_name, direction="input").inc(
                input_tokens
            )
            self._counters["tokens_used"].labels(agent_name=agent_name, direction="output").inc(
                output_tokens
            )

    def set_active_sessions(self, agent_name: str, count: int):
        if "active_sessions" in self._gauges:
            self._gauges["active_sessions"].labels(agent_name=agent_name).set(count)



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime


@dataclass
class ConversationTurn:
    role: str
    content: str
    timestamp: datetime
    metadata: dict = field(default_factory=dict)
    token_count: int = 0


@dataclass
class SemanticMemory:
    content: str
    embedding: list[float]
    metadata: dict
    similarity: float = 0.0
    source: str = ""


@dataclass
class EpisodicMemory:
    event_type: str
    summary: str
    context: dict
    outcome: dict
    timestamp: datetime
    importance_score: float = 0.5


class ConversationBackend(ABC):
    @abstractmethod
    async def save_turn(self, session_id, turn): ...

    @abstractmethod
    async def get_history(self, session_id, limit=50) -> list[ConversationTurn]: ...

    @abstractmethod
    async def clear(self, session_id): ...

    @abstractmethod
    async def get_summary(self, session_id) -> str | None: ...

    @abstractmethod
    async def save_summary(self, session_id, summary): ...


class SemanticBackend(ABC):
    @abstractmethod
    async def store(self, agent_id, content, embedding, metadata): ...

    @abstractmethod
    async def search(
        self, agent_id, query_embedding, top_k=10, threshold=0.7
    ) -> list[SemanticMemory]: ...

    @abstractmethod
    async def delete(self, agent_id, memory_id): ...


class EpisodicBackend(ABC):
    @abstractmethod
    async def record(self, agent_id, episode): ...

    @abstractmethod
    async def recall(
        self, agent_id, event_type=None, since=None, limit=20
    ) -> list[EpisodicMemory]: ...


class MemoryManager:
    def __init__(
        self,
        agent_id,
        config,
        conversation=None,
        semantic=None,
        episodic=None,
        embedding_fn=None,
        summarize_fn=None,
    ):
        self.agent_id = agent_id
        self.config = config
        self._conversation = conversation
        self._semantic = semantic
        self._episodic = episodic
        self._embed = embedding_fn
        self._summarize = summarize_fn

    async def build_context(self, session_id, current_query, max_tokens=4096):
        context = {"conversation": [], "semantic": [], "episodic": []}
        token_budget = max_tokens

        if self._conversation:
            strategy = self.config.get("conversational", {}).get("strategy", "sliding_window")
            if strategy == "sliding_window":
                turns = await self._conversation.get_history(session_id)
                context["conversation"] = turns
                token_budget -= sum(t.token_count for t in turns)
            elif strategy == "summary":
                summary = await self._conversation.get_summary(session_id)
                recent = await self._conversation.get_history(session_id, limit=5)
                context["conversation_summary"] = summary
                context["conversation"] = recent
            elif strategy == "token_budget":
                turns = await self._conversation.get_history(session_id)
                selected, used = [], 0
                for turn in reversed(turns):
                    if used + turn.token_count > token_budget * 0.5:
                        break
                    selected.insert(0, turn)
                    used += turn.token_count
                context["conversation"] = selected
                token_budget -= used

        if self._semantic and self._embed:
            query_emb = await self._embed(current_query)
            threshold = self.config.get("semantic", {}).get("similarity_threshold", 0.75)
            max_results = self.config.get("semantic", {}).get("max_results", 10)
            context["semantic"] = await self._semantic.search(
                self.agent_id, query_emb, top_k=max_results, threshold=threshold
            )

        if self._episodic:
            context["episodic"] = await self._episodic.recall(self.agent_id, limit=5)

        return context

    async def clear_history(self, session_id):
        if self._conversation:
            await self._conversation.clear(session_id)

    async def persist_turn(self, session_id, turn):
        if self._conversation:
            await self._conversation.save_turn(session_id, turn)
            history = await self._conversation.get_history(session_id)
            max_turns = self.config.get("conversational", {}).get("max_turns", 50)
            if len(history) > max_turns and self._summarize:
                summary = await self._summarize(history[:-10])
                await self._conversation.save_summary(session_id, summary)

        if self._semantic and self._embed and turn.role == "assistant" and turn.token_count > 50:
            emb = await self._embed(turn.content)
            await self._semantic.store(
                self.agent_id,
                turn.content,
                emb,
                {
                    "session_id": session_id,
                    "timestamp": turn.timestamp.isoformat(),
                },
            )



"""Mesh state data models."""

from __future__ import annotations

from dataclasses import dataclass, field


@dataclass
class NodeLoad:
    """Resource usage metrics for a node."""

    cpu_percent: float = 0.0
    memory_percent: float = 0.0
    active_requests: int = 0

    def to_dict(self) -> dict:
        return {
            "cpu_percent": self.cpu_percent,
            "memory_percent": self.memory_percent,
            "active_requests": self.active_requests,
        }

    @classmethod
    def from_dict(cls, data: dict) -> NodeLoad:
        return cls(
            cpu_percent=data.get("cpu_percent", 0.0),
            memory_percent=data.get("memory_percent", 0.0),
            active_requests=data.get("active_requests", 0),
        )


@dataclass
class NodeState:
    """State of a single node in the mesh."""

    node_id: str
    name: str
    url: str
    services: list[str]
    agents: list[str]
    load: NodeLoad
    joined_at: float
    last_heartbeat: float
    leader: bool = False
    status: str = "alive"

    def to_dict(self) -> dict:
        return {
            "node_id": self.node_id,
            "name": self.name,
            "url": self.url,
            "services": self.services,
            "agents": self.agents,
            "load": self.load.to_dict(),
            "leader": self.leader,
            "joined_at": self.joined_at,
            "last_heartbeat": self.last_heartbeat,
            "status": self.status,
        }

    @classmethod
    def from_dict(cls, data: dict) -> NodeState:
        return cls(
            node_id=data["node_id"],
            name=data["name"],
            url=data["url"],
            services=data.get("services", []),
            agents=data.get("agents", []),
            load=NodeLoad.from_dict(data.get("load", {})),
            leader=data.get("leader", False),
            joined_at=data["joined_at"],
            last_heartbeat=data["last_heartbeat"],
            status=data.get("status", "alive"),
        )


@dataclass
class ClusterState:
    """State of the entire mesh cluster."""

    nodes: dict[str, NodeState] = field(default_factory=dict)
    leader_id: str | None = None
    version: int = 0

    def add_node(self, node: NodeState) -> None:
        self.nodes[node.node_id] = node
        self.version += 1

    def remove_node(self, node_id: str) -> None:
        self.nodes.pop(node_id, None)
        self.version += 1

    def merge(self, incoming: list[NodeState]) -> None:
        for node in incoming:
            existing = self.nodes.get(node.node_id)
            if not existing or node.last_heartbeat > existing.last_heartbeat:
                self.nodes[node.node_id] = node
                self.version += 1

    def alive_nodes(self) -> list[NodeState]:
        return [n for n in self.nodes.values() if n.status == "alive"]

    def to_dict(self) -> dict:
        return {
            "nodes": [n.to_dict() for n in self.nodes.values()],
            "leader_id": self.leader_id,
            "version": self.version,
        }



from dataclasses import dataclass
from typing import Any


@dataclass
class TorchServingConfig:
    model_path: str
    device: str = "cpu"
    dtype: str = "float32"


class TorchModelServer:
    """Serve PyTorch models."""

    def __init__(self, config: TorchServingConfig):
        self._config = config
        self._model = None

    async def load(self):
        try:
            import torch

            self._model = torch.jit.load(self._config.model_path)
            device = torch.device(self._config.device)
            self._model = self._model.to(device)
            self._model.eval()
        except ImportError:
            raise RuntimeError("torch not installed")

    async def predict(self, inputs: Any) -> Any:
        if not self._model:
            await self.load()
        import torch

        with torch.no_grad():
            return self._model(inputs)



"""TerraformRunner — thin async wrapper around the tofu/terraform CLI binary."""

from __future__ import annotations

import asyncio
import json
import os
import re
import shutil
from pathlib import Path

from astromesh_orbit.core.provider import ApplyResult, PlanResult


def _resolve_binary() -> str:
    """Resolve the IaC binary: TOFU_PATH env > tofu in PATH > terraform in PATH > 'tofu'."""
    tofu_path = os.environ.get("TOFU_PATH")
    if tofu_path and os.path.isfile(tofu_path):
        return tofu_path
    found = shutil.which("tofu") or shutil.which("terraform")
    if found:
        return found
    return "tofu"


class TerraformNotFoundError(Exception):
    """Raised when neither tofu nor terraform binary is found."""

    def __init__(self) -> None:
        super().__init__(
            "OpenTofu/Terraform is not installed or not in PATH.\n"
            "Install OpenTofu from: https://opentofu.org/docs/intro/install/"
        )


class TerraformRunner:
    """Async subprocess wrapper for the tofu/terraform CLI."""

    def __init__(self) -> None:
        self._bin = _resolve_binary()

    async def _run(self, args: list[str], cwd: Path | None = None) -> tuple[int, str, str]:
        proc = await asyncio.create_subprocess_exec(
            *args,
            cwd=cwd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        stdout, stderr = await proc.communicate()
        return proc.returncode or 0, stdout.decode(), stderr.decode()

    async def check_installed(self) -> str:
        """Check tofu/terraform is installed. Returns version string."""
        try:
            code, stdout, _ = await self._run([self._bin, "version", "-json"])
        except FileNotFoundError:
            raise TerraformNotFoundError()
        if code != 0:
            raise TerraformNotFoundError()
        try:
            data = json.loads(stdout)
            return data.get("terraform_version", data.get("tofu_version", "unknown"))
        except json.JSONDecodeError:
            match = re.search(r"(?:Terraform|OpenTofu) v(\S+)", stdout)
            return match.group(1) if match else "unknown"

    async def init(self, work_dir: Path) -> None:
        code, stdout, stderr = await self._run(
            [self._bin, "init", "-input=false", "-no-color"], cwd=work_dir
        )
        if code != 0:
            raise RuntimeError(f"init failed:\n{stderr or stdout}")

    async def plan(self, work_dir: Path) -> PlanResult:
        code, stdout, stderr = await self._run(
            [self._bin, "plan", "-input=false", "-no-color"], cwd=work_dir
        )
        combined = stdout + stderr
        create = re.findall(r"\+ (\S+)", combined)
        update = re.findall(r"~ (\S+)", combined)
        destroy = re.findall(r"- (\S+)", combined)
        return PlanResult(
            resources_to_create=create,
            resources_to_update=update,
            resources_to_destroy=destroy,
            raw_output=combined,
        )

    async def apply(self, work_dir: Path, auto_approve: bool = False) -> ApplyResult:
        args = [self._bin, "apply", "-input=false", "-no-color"]
        if auto_approve:
            args.append("-auto-approve")
        code, stdout, stderr = await self._run(args, cwd=work_dir)
        if code != 0:
            return ApplyResult(success=False, outputs={}, raw_output=stderr or stdout)
        outputs = await self.output(work_dir)
        return ApplyResult(success=True, outputs=outputs, raw_output=stdout)

    async def destroy(self, work_dir: Path, auto_approve: bool = False) -> None:
        args = [self._bin, "destroy", "-input=false", "-no-color"]
        if auto_approve:
            args.append("-auto-approve")
        code, stdout, stderr = await self._run(args, cwd=work_dir)
        if code != 0:
            raise RuntimeError(f"destroy failed:\n{stderr or stdout}")

    async def output(self, work_dir: Path) -> dict[str, str]:
        code, stdout, _ = await self._run(
            [self._bin, "output", "-json", "-no-color"], cwd=work_dir
        )
        if code != 0 or not stdout.strip():
            return {}
        data = json.loads(stdout)
        return {k: v["value"] for k, v in data.items() if isinstance(v, dict) and "value" in v}



from astromesh.ml.model_registry import ModelRegistry, ModelFormat, ModelStatus
from astromesh.ml.serving.onnx_runtime import ONNXModelServer, ONNXServingConfig
from astromesh.ml.serving.torch_serve import TorchModelServer, TorchServingConfig
from astromesh.ml.training.classifier import ClassifierTrainer, ClassifierConfig
from astromesh.ml.training.embeddings import EmbeddingTrainer, EmbeddingTrainerConfig


def test_register_model():
    registry = ModelRegistry()
    info = registry.register("my-classifier", "1.0", ModelFormat.ONNX,
                             "/models/classifier.onnx", "classification")
    assert info.name == "my-classifier"
    assert info.status == ModelStatus.REGISTERED


def test_get_model():
    registry = ModelRegistry()
    registry.register("model", "1.0", ModelFormat.PYTORCH, "/path", "generation")
    registry.register("model", "2.0", ModelFormat.PYTORCH, "/path2", "generation")
    info = registry.get("model", "latest")
    assert info.version == "2.0"


def test_list_models_by_task():
    registry = ModelRegistry()
    registry.register("cls1", "1.0", ModelFormat.ONNX, "/p1", "classification")
    registry.register("emb1", "1.0", ModelFormat.ONNX, "/p2", "embedding")
    cls_models = registry.list_models(task="classification")
    assert len(cls_models) == 1
    assert cls_models[0].name == "cls1"


def test_unregister_model():
    registry = ModelRegistry()
    registry.register("temp", "1.0", ModelFormat.ONNX, "/p", "test")
    registry.unregister("temp", "1.0")
    assert registry.get("temp", "1.0") is None


async def test_classifier_trainer():
    config = ClassifierConfig(model_name="test-model", epochs=1)
    trainer = ClassifierTrainer(config)
    await trainer.prepare([{"text": "hello", "label": 0}])
    metrics = await trainer.train()
    assert metrics["status"] == "training_complete"


async def test_embedding_trainer():
    config = EmbeddingTrainerConfig(base_model="test-model", epochs=1)
    trainer = EmbeddingTrainer(config)
    await trainer.prepare([{"query": "q", "positive": "p"}])
    metrics = await trainer.train()
    assert metrics["status"] == "training_complete"


def test_onnx_serving_config():
    config = ONNXServingConfig(model_path="/models/test.onnx", device="cpu")
    server = ONNXModelServer(config)
    assert server._config.device == "cpu"


def test_torch_serving_config():
    config = TorchServingConfig(model_path="/models/test.pt")
    server = TorchModelServer(config)
    assert server._config.device == "cpu"



"""Tests for platform auto-detection."""

from unittest.mock import patch

from astromesh_node.platform.base import UnsupportedPlatformError
from astromesh_node.platform.detect import get_service_manager
from astromesh_node.platform.foreground import ForegroundManager

import pytest


def test_foreground_flag_returns_foreground_manager():
    mgr = get_service_manager(foreground=True)
    assert isinstance(mgr, ForegroundManager)


def test_unsupported_platform_raises():
    with patch("astromesh_node.platform.detect.sys") as mock_sys:
        mock_sys.platform = "freebsd"
        with pytest.raises(UnsupportedPlatformError, match="freebsd"):
            get_service_manager(foreground=False)


def test_linux_returns_systemd_manager():
    with patch("astromesh_node.platform.detect.sys") as mock_sys:
        mock_sys.platform = "linux"
        mgr = get_service_manager(foreground=False)
        assert type(mgr).__name__ == "SystemdManager"


def test_darwin_returns_launchd_manager():
    with patch("astromesh_node.platform.detect.sys") as mock_sys:
        mock_sys.platform = "darwin"
        mgr = get_service_manager(foreground=False)
        assert type(mgr).__name__ == "LaunchdManager"


def test_win32_returns_windows_manager():
    with patch("astromesh_node.platform.detect.sys") as mock_sys:
        mock_sys.platform = "win32"
        mgr = get_service_manager(foreground=False)
        assert type(mgr).__name__ == "WindowsServiceManager"



"""Tests for GCPProvider — template generation and protocol conformance."""

from pathlib import Path
from unittest.mock import AsyncMock, patch

import pytest

from astromesh_orbit.config import OrbitConfig
from astromesh_orbit.core.provider import (
    ApplyResult,
    CheckResult,
    OrbitProvider,
    ValidationResult,
)
from astromesh_orbit.providers.gcp.provider import GCPProvider


@pytest.fixture
def provider():
    return GCPProvider()


@pytest.fixture
def config(sample_orbit_yaml: Path) -> OrbitConfig:
    return OrbitConfig.from_yaml(sample_orbit_yaml)


def test_gcp_provider_implements_protocol():
    """GCPProvider must satisfy OrbitProvider Protocol."""
    assert isinstance(GCPProvider(), OrbitProvider)


def test_gcp_provider_name():
    assert GCPProvider().name == "gcp"


async def test_generate_creates_tf_files(
    provider: GCPProvider, config: OrbitConfig, output_dir: Path
):
    files = await provider.generate(config, output_dir)
    assert len(files) >= 8
    names = [f.name for f in files]
    assert "main.tf" in names
    assert "cloud_run.tf" in names
    assert "cloud_sql.tf" in names
    assert "memorystore.tf" in names
    assert "secrets.tf" in names
    assert "iam.tf" in names
    assert "networking.tf" in names
    assert "outputs.tf" in names


async def test_generate_tf_files_are_valid_hcl(
    provider: GCPProvider, config: OrbitConfig, output_dir: Path
):
    """Generated files should contain valid-looking HCL (basic structural check)."""
    files = await provider.generate(config, output_dir)
    for f in files:
        content = f.read_text()
        assert len(content) > 0, f"{f.name} is empty"
        # Every .tf file should have at least one block
        assert any(
            kw in content for kw in ["resource", "variable", "output", "provider", "terraform"]
        ), f"{f.name} has no HCL blocks"


async def test_generate_includes_project_in_main_tf(
    provider: GCPProvider, config: OrbitConfig, output_dir: Path
):
    await provider.generate(config, output_dir)
    main_tf = (output_dir / "main.tf").read_text()
    assert "test-project-123" in main_tf


async def test_eject_writes_to_output_dir(
    provider: GCPProvider, config: OrbitConfig, tmp_path: Path
):
    eject_dir = tmp_path / "ejected"
    result = await provider.eject(config, eject_dir)
    assert result == eject_dir
    assert (eject_dir / "main.tf").exists()
    # Ejected files should have comments
    content = (eject_dir / "main.tf").read_text()
    assert "#" in content  # Should have explanatory comments
    # Should produce terraform.tfvars with resolved values
    tfvars = (eject_dir / "terraform.tfvars").read_text()
    assert "test-project-123" in tfvars
    assert "us-central1" in tfvars


async def test_validate_calls_all_checks(provider: GCPProvider, config: OrbitConfig):
    """Validate should run auth, project, and API checks."""
    with (
        patch(
            "astromesh_orbit.providers.gcp.provider.check_gcloud_auth", new_callable=AsyncMock
        ) as auth,
        patch(
            "astromesh_orbit.providers.gcp.provider.check_project", new_callable=AsyncMock
        ) as proj,
        patch(
            "astromesh_orbit.providers.gcp.provider.check_apis_enabled", new_callable=AsyncMock
        ) as apis,
    ):
        auth.return_value = CheckResult(name="auth", passed=True, message="ok")
        proj.return_value = CheckResult(name="project", passed=True, message="ok")
        apis.return_value = [CheckResult(name="api", passed=True, message="ok")]
        result = await provider.validate(config)
        assert result.ok is True
        auth.assert_called_once()
        proj.assert_called_once_with("test-project-123")


async def test_validate_fails_on_missing_auth(provider: GCPProvider, config: OrbitConfig):
    with (
        patch(
            "astromesh_orbit.providers.gcp.provider.check_gcloud_auth", new_callable=AsyncMock
        ) as auth,
        patch(
            "astromesh_orbit.providers.gcp.provider.check_project", new_callable=AsyncMock
        ) as proj,
        patch(
            "astromesh_orbit.providers.gcp.provider.check_apis_enabled", new_callable=AsyncMock
        ) as apis,
    ):
        auth.return_value = CheckResult(
            name="auth", passed=False, message="not authed", remediation="gcloud auth login"
        )
        proj.return_value = CheckResult(name="project", passed=True, message="ok")
        apis.return_value = []
        result = await provider.validate(config)
        assert result.ok is False


async def test_provision_orchestrates_full_flow(
    provider: GCPProvider, config: OrbitConfig, output_dir: Path
):
    """Provision should: validate -> ensure bucket -> generate -> init -> apply -> write env."""
    with (
        patch.object(provider, "validate", new_callable=AsyncMock) as mock_val,
        patch.object(provider, "generate", new_callable=AsyncMock) as mock_gen,
        patch(
            "astromesh_orbit.providers.gcp.provider.ensure_gcs_state_bucket",
            new_callable=AsyncMock,
        ) as mock_bucket,
        patch.object(provider._tf, "init", new_callable=AsyncMock),
        patch.object(provider._tf, "apply", new_callable=AsyncMock) as mock_apply,
    ):
        mock_val.return_value = ValidationResult(ok=True, checks=[])
        mock_gen.return_value = [output_dir / "main.tf"]
        mock_bucket.return_value = "test-bucket"
        mock_apply.return_value = ApplyResult(
            success=True,
            outputs={
                "runtime_url": "https://rt.run.app",
            },
            raw_output="Apply complete!",
        )
        result = await provider.provision(config, output_dir)
        assert result.endpoints["runtime"] == "https://rt.run.app"
        assert result.env_file.name == "orbit.env"
        mock_val.assert_called_once()
        mock_bucket.assert_called_once()



"""Memory configuration normalization."""

from __future__ import annotations

CONVERSATIONAL_DEFAULTS = {
    "strategy": "sliding_window",
    "max_turns": 50,
}

SEMANTIC_DEFAULTS = {
    "similarity_threshold": 0.7,
    "max_results": 5,
}


def normalize_memory_config(memory: str | dict | None) -> dict:
    """Normalize memory config from shorthand or full dict.

    Accepts:
        - None → {}
        - "sqlite" → {"conversational": {"backend": "sqlite", ...defaults}}
        - {"conversational": "redis"} → {"conversational": {"backend": "redis", ...defaults}}
        - {"conversational": {"backend": "sqlite", "strategy": "summary"}} → passthrough with defaults
    """
    if memory is None:
        return {}

    if isinstance(memory, str):
        return {
            "conversational": {"backend": memory, **CONVERSATIONAL_DEFAULTS},
        }

    result = {}
    for mem_type, value in memory.items():
        if isinstance(value, str):
            defaults = CONVERSATIONAL_DEFAULTS if mem_type == "conversational" else SEMANTIC_DEFAULTS
            result[mem_type] = {"backend": value, **defaults}
        elif isinstance(value, dict):
            defaults = CONVERSATIONAL_DEFAULTS if mem_type == "conversational" else SEMANTIC_DEFAULTS
            merged = {**defaults, **value}
            result[mem_type] = merged
        else:
            result[mem_type] = value

    return result



from dataclasses import dataclass
from typing import Any


@dataclass
class ONNXServingConfig:
    model_path: str
    device: str = "cpu"
    num_threads: int = 4
    optimization_level: int = 1


class ONNXModelServer:
    """Serve ONNX models with optimized inference."""

    def __init__(self, config: ONNXServingConfig):
        self._config = config
        self._session = None

    async def load(self):
        try:
            import onnxruntime as ort

            opts = ort.SessionOptions()
            opts.intra_op_num_threads = self._config.num_threads
            opts.graph_optimization_level = ort.GraphOptimizationLevel(
                self._config.optimization_level
            )
            providers = ["CPUExecutionProvider"]
            if self._config.device == "cuda":
                providers.insert(0, "CUDAExecutionProvider")
            self._session = ort.InferenceSession(self._config.model_path, opts, providers=providers)
        except ImportError:
            raise RuntimeError("onnxruntime not installed")

    async def predict(self, inputs: dict) -> Any:
        if not self._session:
            await self.load()
        return self._session.run(None, inputs)

    def get_input_names(self) -> list[str]:
        if self._session:
            return [inp.name for inp in self._session.get_inputs()]
        return []

    def get_output_names(self) -> list[str]:
        if self._session:
            return [out.name for out in self._session.get_outputs()]
        return []



from asgi_lifespan import LifespanManager
from httpx import ASGITransport, AsyncClient

from astromesh.api.main import app


async def test_cors_preflight():
    async with LifespanManager(app):
        async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as client:
            resp = await client.options(
                "/v1/health",
                headers={
                    "Origin": "http://localhost:5173",
                    "Access-Control-Request-Method": "GET",
                },
            )
    assert resp.status_code == 200
    assert "access-control-allow-origin" in resp.headers



import pytest
from dataclasses import dataclass
from unittest.mock import AsyncMock

from astromesh.orchestration.patterns import ReActPattern


@dataclass
class CompletionResponse:
    content: str
    tool_calls: list | None = None


def make_response(content, tool_calls=None):
    return CompletionResponse(content=content, tool_calls=tool_calls)


@pytest.mark.asyncio
async def test_react_direct_answer():
    """No tool calls -> final answer returned immediately."""
    model_fn = AsyncMock(return_value=make_response("The answer is 42."))
    tool_fn = AsyncMock()

    pattern = ReActPattern()
    result = await pattern.execute(
        query="What is the answer?",
        context={},
        model_fn=model_fn,
        tool_fn=tool_fn,
        tools=[],
    )

    assert result["answer"] == "The answer is 42."
    assert len(result["steps"]) == 1
    assert result["steps"][0].result == "The answer is 42."
    tool_fn.assert_not_called()


@pytest.mark.asyncio
async def test_react_with_tool_call():
    """One tool call followed by a final answer."""
    tool_call = {"id": "tc_1", "name": "lookup", "arguments": {"query": "weather"}}

    # First call returns a tool call; second call returns the final answer
    model_fn = AsyncMock(
        side_effect=[
            make_response("Let me look that up.", tool_calls=[tool_call]),
            make_response("It is sunny today."),
        ]
    )
    tool_fn = AsyncMock(return_value={"weather": "sunny"})

    pattern = ReActPattern()
    result = await pattern.execute(
        query="What is the weather?",
        context={},
        model_fn=model_fn,
        tool_fn=tool_fn,
        tools=[],
    )

    assert result["answer"] == "It is sunny today."
    assert len(result["steps"]) == 2
    # First step is the tool call
    assert result["steps"][0].action == "lookup"
    assert result["steps"][0].observation == str({"weather": "sunny"})
    # Second step is the final answer
    assert result["steps"][1].result == "It is sunny today."
    tool_fn.assert_called_once_with("lookup", {"query": "weather"})


@pytest.mark.asyncio
async def test_react_max_iterations():
    """Always returns tool_calls -> hits max iterations."""
    tool_call = {"id": "tc_loop", "name": "search", "arguments": {"q": "info"}}
    model_fn = AsyncMock(
        return_value=make_response("Thinking...", tool_calls=[tool_call])
    )
    tool_fn = AsyncMock(return_value={"result": "partial"})

    pattern = ReActPattern()
    result = await pattern.execute(
        query="Find info",
        context={},
        model_fn=model_fn,
        tool_fn=tool_fn,
        tools=[],
        max_iterations=3,
    )

    assert result["answer"] == "Max iterations reached"
    assert len(result["steps"]) == 3
    assert all(step.action == "search" for step in result["steps"])


@pytest.mark.asyncio
async def test_plan_execute_pattern():
    from astromesh.orchestration.patterns import PlanAndExecutePattern
    plan_response = make_response('{"steps": [{"step": 1, "description": "Search", "tool": null, "depends_on": []}]}')
    step_response = make_response("Step 1 result")
    final_response = make_response("Final synthesized answer")
    model_fn = AsyncMock(side_effect=[plan_response, step_response, final_response])
    tool_fn = AsyncMock()
    pattern = PlanAndExecutePattern()
    result = await pattern.execute("Do research", {}, model_fn, tool_fn, [])
    assert result["answer"] == "Final synthesized answer"


@pytest.mark.asyncio
async def test_parallel_fan_out():
    from astromesh.orchestration.patterns import ParallelFanOutPattern
    decompose = make_response('["subtask 1", "subtask 2"]')
    sub1 = make_response("result 1")
    sub2 = make_response("result 2")
    final = make_response("Aggregated answer")
    model_fn = AsyncMock(side_effect=[decompose, sub1, sub2, final])
    pattern = ParallelFanOutPattern()
    result = await pattern.execute("Do things", {}, model_fn, AsyncMock(), [])
    assert result["answer"] == "Aggregated answer"


@pytest.mark.asyncio
async def test_pipeline_pattern():
    from astromesh.orchestration.patterns import PipelinePattern
    r1 = make_response("analyzed")
    r2 = make_response("processed")
    r3 = make_response("synthesized")
    model_fn = AsyncMock(side_effect=[r1, r2, r3])
    pattern = PipelinePattern()
    result = await pattern.execute("input", {}, model_fn, AsyncMock(), [])
    assert result["answer"] == "synthesized"


@pytest.mark.asyncio
async def test_supervisor_pattern():
    from astromesh.orchestration.supervisor import SupervisorPattern
    resp = make_response('{"final_answer": "done"}')
    model_fn = AsyncMock(return_value=resp)
    pattern = SupervisorPattern()
    result = await pattern.execute("task", {}, model_fn, AsyncMock(), [])
    assert result["answer"] == "done"


@pytest.mark.asyncio
async def test_swarm_pattern():
    from astromesh.orchestration.swarm import SwarmPattern
    resp = make_response("The answer is 42")
    model_fn = AsyncMock(return_value=resp)
    pattern = SwarmPattern()
    result = await pattern.execute("question", {}, model_fn, AsyncMock(), [])
    assert result["answer"] == "The answer is 42"



"""Service manager for Astromesh Node node roles."""

DEFAULT_SERVICES = (
    "api",
    "agents",
    "inference",
    "memory",
    "tools",
    "channels",
    "rag",
    "observability",
)


class ServiceManager:
    """Controls which services are active on this node."""

    def __init__(self, services_config: dict[str, bool]):
        self._services: dict[str, bool] = {}
        for service in DEFAULT_SERVICES:
            self._services[service] = services_config.get(service, True)

    def is_enabled(self, service: str) -> bool:
        return self._services.get(service, False)

    def enabled_services(self) -> list[str]:
        return [s for s, enabled in self._services.items() if enabled]

    def validate(self) -> list[str]:
        warnings: list[str] = []
        if self._services.get("agents") and not self._services.get("tools"):
            warnings.append("agents enabled without tools — agents won't be able to use tools")
        if self._services.get("agents") and not self._services.get("memory"):
            warnings.append(
                "agents enabled without memory — agents won't have conversation history"
            )
        return warnings

    def to_dict(self) -> dict[str, bool]:
        return dict(self._services)



from astromesh_adk.mcp import mcp_tools, MCPToolSet


def test_mcp_tools_returns_toolset():
    toolset = mcp_tools(transport="stdio", command="echo", args=["hello"])
    assert isinstance(toolset, MCPToolSet)


def test_mcp_toolset_is_iterable():
    toolset = mcp_tools(transport="stdio", command="echo", args=[])
    # Before discovery, iteration yields the toolset itself as a lazy marker
    items = list(toolset)
    assert len(items) == 1
    assert items[0] is toolset


def test_mcp_toolset_stores_config():
    toolset = mcp_tools(
        transport="http",
        command=None,
        args=[],
        url="https://mcp.example.com",
        headers={"Authorization": "Bearer xxx"},
    )
    assert toolset.config["transport"] == "http"
    assert toolset.config["url"] == "https://mcp.example.com"


def test_mcp_toolset_discovered_flag():
    toolset = mcp_tools(transport="stdio", command="echo", args=[])
    assert toolset.discovered is False



"""Shared fixtures for Orbit tests."""

from pathlib import Path
import pytest
import yaml


SAMPLE_ORBIT_YAML = {
    "apiVersion": "astromesh/v1",
    "kind": "OrbitDeployment",
    "metadata": {"name": "test-astromesh", "environment": "dev"},
    "spec": {
        "provider": {"name": "gcp", "project": "test-project-123", "region": "us-central1"},
        "compute": {
            "runtime": {"min_instances": 1, "max_instances": 3, "cpu": "2", "memory": "2Gi"},
        },
        "database": {
            "tier": "db-f1-micro",
            "version": "POSTGRES_16",
            "storage_gb": 10,
            "high_availability": False,
        },
        "cache": {"tier": "basic", "memory_gb": 1},
        "secrets": {"provider_keys": True, "jwt_secret": True},
        "images": {
            "runtime": "fulfarodev/astromesh:latest",
        },
    },
}


@pytest.fixture
def sample_orbit_dict():
    import copy

    return copy.deepcopy(SAMPLE_ORBIT_YAML)


@pytest.fixture
def sample_orbit_yaml(tmp_path: Path) -> Path:
    path = tmp_path / "orbit.yaml"
    path.write_text(yaml.dump(SAMPLE_ORBIT_YAML, default_flow_style=False))
    return path


@pytest.fixture
def output_dir(tmp_path: Path) -> Path:
    d = tmp_path / "generated"
    d.mkdir()
    return d



"""GCP pre-deploy validation helpers.

Supports two authentication methods:
  1. gcloud CLI — used when GOOGLE_APPLICATION_CREDENTIALS is not set
  2. Service Account key — used when GOOGLE_APPLICATION_CREDENTIALS points to a JSON key file
"""

from __future__ import annotations

import asyncio
import json
import os
import sys

from astromesh_orbit.core.provider import CheckResult

REQUIRED_APIS = [
    "run.googleapis.com",
    "sqladmin.googleapis.com",
    "redis.googleapis.com",
    "secretmanager.googleapis.com",
    "vpcaccess.googleapis.com",
]

# Windows needs gcloud.cmd; create_subprocess_exec can't resolve .cmd from PATH
_GCLOUD = "gcloud.cmd" if sys.platform == "win32" else "gcloud"


def _has_service_account_key() -> str | None:
    """Return the path if GOOGLE_APPLICATION_CREDENTIALS is set and the file exists."""
    path = os.environ.get("GOOGLE_APPLICATION_CREDENTIALS")
    if path and os.path.isfile(path):
        return path
    return None


# ---------------------------------------------------------------------------
# gcloud CLI helpers
# ---------------------------------------------------------------------------

async def _run_gcloud(*args: str) -> tuple[int, str, str]:
    try:
        proc = await asyncio.create_subprocess_exec(
            _GCLOUD,
            *args,
            "--format=json",
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
    except FileNotFoundError:
        return 127, "", "gcloud CLI not found in PATH"
    stdout, stderr = await proc.communicate()
    return proc.returncode or 0, stdout.decode(), stderr.decode()


# ---------------------------------------------------------------------------
# Service-account (google-auth library) helpers
# ---------------------------------------------------------------------------

def _load_sa_credentials():
    """Load default credentials from GOOGLE_APPLICATION_CREDENTIALS."""
    from google.auth import default  # noqa: lazy import – optional dep
    from google.auth.transport.requests import Request

    creds, project = default(
        scopes=["https://www.googleapis.com/auth/cloud-platform"],
    )
    creds.refresh(Request())
    return creds, project


def _sa_check_auth() -> CheckResult:
    """Validate service-account credentials can be loaded."""
    key_path = _has_service_account_key()
    try:
        creds, _ = _load_sa_credentials()
        email = getattr(creds, "service_account_email", None) or "service-account"
        return CheckResult(
            name="gcloud_auth",
            passed=True,
            message=f"Authenticated via service account: {email}",
        )
    except Exception as exc:
        return CheckResult(
            name="gcloud_auth",
            passed=False,
            message=f"Service account auth failed: {exc}",
            remediation=f"Check GOOGLE_APPLICATION_CREDENTIALS={key_path}",
        )


def _sa_check_project(project: str) -> CheckResult:
    """Validate project exists using google-cloud-resource-manager."""
    try:
        from google.cloud.resourcemanager_v3 import ProjectsClient

        client = ProjectsClient()
        p = client.get_project(name=f"projects/{project}")
        return CheckResult(
            name="project_exists",
            passed=True,
            message=f"Project {p.project_id} found",
        )
    except Exception as exc:
        return CheckResult(
            name="project_exists",
            passed=False,
            message=f"Project {project} not found or no access: {exc}",
            remediation=f"Verify the project ID and service account permissions",
        )


def _sa_check_apis_enabled(project: str) -> list[CheckResult]:
    """Check enabled APIs using the Service Usage REST API via google-auth."""
    try:
        from google.auth.transport.requests import AuthorizedSession

        creds, _ = _load_sa_credentials()
        session = AuthorizedSession(creds)
        url = (
            f"https://serviceusage.googleapis.com/v1/projects/{project}"
            f"/services?filter=state:ENABLED&fields=services/config/name"
        )
        resp = session.get(url)
        resp.raise_for_status()
        data = resp.json()
        enabled = {svc.get("config", {}).get("name", "") for svc in data.get("services", [])}
    except Exception:
        # If we can't query Service Usage, mark all as unknown/failed
        return [
            CheckResult(
                name=f"api_{api}",
                passed=False,
                message=f"{api} — unable to verify (Service Usage API may not be enabled)",
                remediation=f"gcloud services enable serviceusage.googleapis.com --project={project}",
            )
            for api in REQUIRED_APIS
        ]

    results = []
    for api in REQUIRED_APIS:
        if api in enabled:
            results.append(CheckResult(name=f"api_{api}", passed=True, message=f"{api} enabled"))
        else:
            results.append(
                CheckResult(
                    name=f"api_{api}",
                    passed=False,
                    message=f"{api} not enabled",
                    remediation=f"gcloud services enable {api} --project={project}",
                )
            )
    return results


# ---------------------------------------------------------------------------
# Public API — auto-selects auth method
# ---------------------------------------------------------------------------

async def check_gcloud_auth() -> CheckResult:
    if _has_service_account_key():
        return _sa_check_auth()

    try:
        code, stdout, _ = await _run_gcloud("auth", "list", "--filter=status:ACTIVE")
        accounts = json.loads(stdout) if stdout.strip() else []
        if code == 0 and accounts:
            return CheckResult(
                name="gcloud_auth",
                passed=True,
                message=f"Authenticated as {accounts[0].get('account', 'unknown')}",
            )
    except (FileNotFoundError, json.JSONDecodeError):
        pass
    return CheckResult(
        name="gcloud_auth",
        passed=False,
        message="gcloud CLI not authenticated and no service account key found",
        remediation="gcloud auth login  OR  set GOOGLE_APPLICATION_CREDENTIALS",
    )


async def check_project(project: str) -> CheckResult:
    if _has_service_account_key():
        return _sa_check_project(project)

    code, _, _ = await _run_gcloud("projects", "describe", project)
    if code == 0:
        return CheckResult(name="project_exists", passed=True, message=f"Project {project} found")
    return CheckResult(
        name="project_exists",
        passed=False,
        message=f"Project {project} not found or no access",
        remediation=f"Verify the project ID and your permissions: gcloud projects describe {project}",
    )


async def check_apis_enabled(project: str) -> list[CheckResult]:
    if _has_service_account_key():
        return _sa_check_apis_enabled(project)

    code, stdout, _ = await _run_gcloud("services", "list", "--enabled", f"--project={project}")
    enabled = set()
    if code == 0 and stdout.strip():
        try:
            for svc in json.loads(stdout):
                name = svc.get("config", {}).get("name", "")
                enabled.add(name)
        except json.JSONDecodeError:
            pass

    results = []
    for api in REQUIRED_APIS:
        if api in enabled:
            results.append(CheckResult(name=f"api_{api}", passed=True, message=f"{api} enabled"))
        else:
            results.append(
                CheckResult(
                    name=f"api_{api}",
                    passed=False,
                    message=f"{api} not enabled",
                    remediation=f"gcloud services enable {api} --project={project}",
                )
            )
    return results



"""astromeshctl — Astromesh CLI management tool."""

import importlib.metadata

import typer

from astromesh_cli import __version__
from astromesh_cli.commands import (
    agents,
    ask,
    dev,
    doctor,
    mesh,
    metrics,
    new,
    peers,
    providers,
    run,
    services,
    status,
    tools,
    traces,
)

app = typer.Typer(
    name="astromeshctl",
    help="Astromesh CLI — manage nodes and clusters.",
    no_args_is_help=True,
)

app.add_typer(status.app, name="status")
app.add_typer(doctor.app, name="doctor")
app.add_typer(agents.app, name="agents")
app.add_typer(providers.app, name="providers")
app.add_typer(mesh.app, name="mesh")
app.add_typer(peers.app, name="peers")
app.add_typer(services.app, name="services")
app.add_typer(new.app, name="new")
app.command("run")(run.run_command)
app.command("dev")(dev.dev_command)
app.add_typer(traces.app, name="traces")
app.add_typer(tools.app, name="tools")
app.command("trace")(traces.trace_command)
app.command("metrics")(metrics.metrics_command)
app.command("cost")(metrics.cost_command)
app.command("ask")(ask.ask_command)


@app.command()
def version():
    """Show astromesh version."""
    typer.echo(f"astromesh {__version__}")


# Plugin discovery — after all static registrations
def _load_plugins(app: typer.Typer) -> None:
    """Discover and register CLI plugins via entry points."""
    try:
        eps = importlib.metadata.entry_points(group="astromeshctl.plugins")
    except TypeError:
        # Python < 3.12 compat
        eps = importlib.metadata.entry_points().get("astromeshctl.plugins", [])
    for ep in eps:
        try:
            register_fn = ep.load()
            register_fn(app)
        except Exception as exc:
            typer.echo(f"Warning: failed to load plugin '{ep.name}': {exc}", err=True)


_load_plugins(app)


if __name__ == "__main__":
    app()



from __future__ import annotations

from astromesh.rag.stores.base import VectorStore


class QdrantStore(VectorStore):
    """Vector store backed by Qdrant."""

    def __init__(
        self,
        collection_name: str = "astromesh",
        url: str = "http://localhost:6333",
        api_key: str | None = None,
        dimensions: int = 384,
    ):
        self.collection_name = collection_name
        self.url = url
        self.api_key = api_key
        self.dimensions = dimensions
        self._client = None

    def _get_client(self):
        if self._client is None:
            from qdrant_client import QdrantClient
            from qdrant_client.models import Distance, VectorParams

            self._client = QdrantClient(url=self.url, api_key=self.api_key)

            # Ensure collection exists
            collections = self._client.get_collections().collections
            if not any(c.name == self.collection_name for c in collections):
                self._client.create_collection(
                    collection_name=self.collection_name,
                    vectors_config=VectorParams(size=self.dimensions, distance=Distance.COSINE),
                )
        return self._client

    async def upsert(self, doc_id: str, embedding: list[float], content: str, metadata: dict):
        from qdrant_client.models import PointStruct

        client = self._get_client()
        point = PointStruct(
            id=doc_id,
            vector=embedding,
            payload={"content": content, **metadata},
        )
        client.upsert(collection_name=self.collection_name, points=[point])

    async def search(
        self,
        query_embedding: list[float],
        top_k: int = 10,
        filters: dict | None = None,
    ) -> list[dict]:
        from qdrant_client.models import FieldCondition, Filter, MatchValue

        client = self._get_client()
        query_filter = None
        if filters:
            conditions = [
                FieldCondition(key=k, match=MatchValue(value=v)) for k, v in filters.items()
            ]
            query_filter = Filter(must=conditions)

        results = client.search(
            collection_name=self.collection_name,
            query_vector=query_embedding,
            limit=top_k,
            query_filter=query_filter,
        )

        return [
            {
                "doc_id": str(hit.id),
                "content": hit.payload.get("content", ""),
                "metadata": {k: v for k, v in hit.payload.items() if k != "content"},
                "score": hit.score,
            }
            for hit in results
        ]

    async def delete(self, doc_id: str):
        from qdrant_client.models import PointIdsList

        client = self._get_client()
        client.delete(
            collection_name=self.collection_name,
            points_selector=PointIdsList(points=[doc_id]),
        )
